#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Set, List, Dict, Iterator, Optional, Iterable
from .graph import Graph
from .rule import Rule

DEBUG_MATCH = False
def match_log(s: str) -> None:
    """Used for debugging the matcher"""
    if DEBUG_MATCH:
        print(s)

class Match:
    dom: Graph
    cod: Graph
    vmap: Dict[int,int]
    vimg: Set[int]
    emap: Dict[int,int]
    eimg: Set[int]

    def __init__(self, dom: Optional[Graph]=None, cod: Optional[Graph]=None, m: Optional[Match]=None) -> None:
        if m:
            self.dom = m.dom
            self.cod = m.cod
            self.vmap = m.vmap.copy()
            self.vimg = m.vimg.copy()
            self.emap = m.emap.copy()
            self.eimg = m.eimg.copy()
        elif dom and cod:
            self.dom = dom
            self.cod = cod
            self.vmap = dict()
            self.vimg = set()
            self.emap = dict()
            self.eimg = set()
        else:
            raise ValueError("Must provide either a match or a pair of graphs")

    def __str__(self) -> str:
        return "  vmap: {}\n  emap: {}".format(str(self.vmap), str(self.emap))

    def copy(self) -> Match:
        return Match(m=self)

    def try_add_vertex(self, v: int, cod_v: int) -> bool:
        match_log("trying to add vertex {} -> {} to match:".format(v, cod_v))
        match_log(str(self))

        # if the vertex is already mapped, only check the new mapping is consistent
        if v in self.vmap:
            match_log("vertex already mapped to {}".format(self.vmap[v]))
            return self.vmap[v] == cod_v

        v_val = self.dom.vertex_data(v).value
        cod_v_val = self.cod.vertex_data(cod_v).value

        if v_val != cod_v_val:
            match_log("vertex failed: values {} != {}".format(v_val, cod_v_val))
            return False

        if self.cod.is_boundary(cod_v) and not self.dom.is_boundary(v):
            match_log("vertex failed: cod v is boundary but dom v is not")
            return False

        # matches are only allowed to be non-injective on the boundary
        if cod_v in self.vimg:
            if not self.dom.is_boundary(v):
                match_log("vertex failed: non-injective on interior vertex")
                return False
            for dv, cv in self.vmap.items():
                if cv == cod_v and not self.dom.is_boundary(dv):
                    match_log("vertex failed: non-injective on interior vertex")
                    return False
        self.vmap[v] = cod_v
        self.vimg.add(cod_v)

        # unless v is a boundary, check that nhd(v) and nhd(vmap(v)) are the same size. Because
        # matchings are required to be injective on edges, this will guarantee that the gluing
        # conditions are satisfied.
        if not self.dom.is_boundary(v):
            if len(self.dom.in_edges(v)) != len(self.cod.in_edges(cod_v)):
                match_log("vertex failed: in_edges cannot satisfy gluing conds")
                return False
            if len(self.dom.out_edges(v)) != len(self.cod.out_edges(cod_v)):
                match_log("vertex failed: out_edges cannot satisfy gluing conds")
                return False

        match_log("vertex success")
        return True

    def try_add_edge(self, e: int, cod_e: int) -> bool:
        match_log("trying to add edge {} -> {} to match:".format(e, cod_e))
        match_log(str(self))

        e_val = self.dom.edge_data(e).value
        cod_e_val = self.cod.edge_data(cod_e).value
        if e_val != cod_e_val:
            match_log("edge failed: values {} != {}".format(e_val, cod_e_val))
            return False

        if cod_e in self.eimg:
            match_log("edge failed: non-injective")
            return False

        self.emap[e] = cod_e
        self.eimg.add(cod_e)

        # compare sources and targets
        s = self.dom.source(e)
        cod_s = self.cod.source(cod_e)
        t = self.dom.target(e)
        cod_t = self.cod.target(cod_e)

        # first the lengths need to be the same
        if len(s) != len(cod_s) or len(t) != len(cod_t):
            match_log("edge failed: source or target len doesn't match image")
            return False

        # then, each vertex that is already mapped needs to be consistent
        for v1, cod_v1 in zip(s + t, cod_s + cod_t):
            if v1 in self.vmap:
                if self.vmap[v1] != cod_v1:
                    match_log("edge failed: inconsistent with previously mapped vertex")
                    return False
            else:
                if not self.try_add_vertex(v1, cod_v1):
                    match_log("edge failed: couldn't add a vertex")
                    return False
        
        match_log("edge success")
        return True

    def dom_nhd_mapped(self, v: int) -> bool:
        """Returns True if nhd(v) is the domain of emap"""
        return (all(e in self.emap for e in self.dom.in_edges(v)) and
                all(e in self.emap for e in self.dom.out_edges(v)))

    # def cod_nhd_mapped(self, cod_v: int):
    #     """Returns True if nhd(cod_v) is the range of emap"""
    #     return (all(e in self.eimg for e in self.cod.in_edges(cod_v)) and
    #             all(e in self.eimg for e in self.cod.out_edges(cod_v)))

    def map_scalars(self) -> bool:
        """Try to extend the match by mapping all scalars (i.e. 0 -> 0 edges)

        Returns True if is successful. Note that any matchings of scalars will yield
        isomorphic results under rewriting, so we don't return a list of all the possible
        matchings.
        """
        cod_sc = []
        for e in self.cod.edges():
            ed = self.cod.edge_data(e)
            if len(ed.s) == 0 and len(ed.t) == 0:
                cod_sc.append((e, ed.value))

        for e in self.dom.edges():
            match_log("trying to map scalar edge {}".format(e))
            ed = self.dom.edge_data(e)
            if len(ed.s) != 0 or len(ed.t) != 0: continue
            found = False
            for i in range(len(cod_sc)):
                e1, val = cod_sc[i]
                if val == ed.value:
                    cod_sc.pop(i)
                    self.emap[e] = e1
                    self.eimg.add(e1)
                    found = True
                    match_log("successfully mapped scalar {} -> {}".format(e, e1))
                    break
            if not found:
                match_log("match failed: could not map scalar edge {}".format(e))
                return False

        return True



    def more(self) -> List[Match]:
        # a list of partial matches the same as this one, but matching 1 more vertex or edge
        ms = []

        # first, try to complete nhds of vertices already matched
        for v in self.vmap:
            # check if a vertex's nhd is already fully mapped
            if self.dom_nhd_mapped(v): continue
            cod_v = self.vmap[v]

            # try to extend the match by mapping the next in_edge
            for e in self.dom.in_edges(v):
                if e in self.emap: continue
                for cod_e in self.cod.in_edges(cod_v):
                    m1 = self.copy()
                    if m1.try_add_edge(e, cod_e):
                        ms.append(m1)
                return ms

            # if there are no unmapped in_edges, try to match the next out_edge
            for e in self.dom.out_edges(v):
                if e in self.emap: continue
                for cod_e in self.cod.out_edges(cod_v):
                    m1 = self.copy()
                    if m1.try_add_edge(e, cod_e):
                        ms.append(m1)
                return ms
            
        # if there are no more vertices with partially-mapped nhds,  try to match a new vertex
        for v in self.dom.vertices():
            # already looked at these
            if v in self.vmap: continue
            
            # create a new match for every possible place v can go
            for cod_v in self.cod.vertices():
                m1 = self.copy()
                if m1.try_add_vertex(v, cod_v):
                    ms.append(m1)
            return ms

        return []
    
    def is_total(self) -> bool:
        return len(self.vmap) == self.dom.num_vertices() and len(self.emap) == self.dom.num_edges()

    def is_surjective(self) -> bool:
        return len(self.vimg) == self.cod.num_vertices() and len(self.eimg) == self.cod.num_edges()

    def is_injective(self) -> bool:
        return len(self.vmap) == len(self.vimg)

    def is_convex(self) -> bool:
        if not self.is_injective():
            return False

        future = self.cod.successors([self.vmap[v] for v in self.dom.outputs() if v in self.vmap])
        for v in self.dom.inputs():
            if v in self.vmap and self.vmap[v] in future:
                return False
        return True



class Matches(Iterable):
    def __init__(self, dom: Graph, cod: Graph, initial_match: Optional[Match] = None, convex: bool=True) -> None:
        if initial_match is None: initial_match = Match(dom=dom, cod=cod) 
        self.convex = convex

        if initial_match.map_scalars():
            self.match_stack = [initial_match]
        else:
            self.match_stack = []

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Match:
        while len(self.match_stack) > 0:
            m = self.match_stack.pop()
            if m.is_total():
                match_log("got successful match:\n" + str(m))
                if self.convex:
                    if m.is_convex():
                        match_log("match is convex, returning")
                        return m
                    else:
                        match_log("match is not convex, dropping")
                else:
                    return m
            else:
                self.match_stack += m.more()
        raise StopIteration

def match_graph(dom: Graph, cod: Graph, convex: bool=True) -> Iterable[Match]:
    return Matches(dom, cod, convex=convex)

def match_rule(r: Rule, g: Graph, convex: bool=True) -> Iterable[Match]:
    return Matches(r.lhs, g, convex=convex)

def find_iso(g: Graph, h: Graph) -> Optional[Match]:
    g_in = g.inputs()
    g_out = g.outputs()
    h_in = h.inputs()
    h_out = h.outputs()
    if len(g_in) != len(h_in) or len(g_out) != len(h_out): return None

    m0 = Match(dom=g, cod=h)
    for i in range(len(g_in)):
        if not m0.try_add_vertex(g_in[i], h_in[i]): return None
    for i in range(len(g_out)):
        if not m0.try_add_vertex(g_out[i], h_out[i]): return None

    for m in Matches(dom=g, cod=h, initial_match=m0, convex=False):
        if m.is_surjective(): return m

    return None



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Set, List, Dict, Iterator, Any, Optional, Iterable, Tuple
from .graph import Graph
from .matcher import Match
from .rule import Rule


def dpo(r: Rule, m: Match) -> Iterable[Match]:
    """Do double-pushout rewriting

    Given a rule r and match of r.lhs into a graph, return a match
    of r.rhs into the rewritten graph.
    """
    # if not r.is_left_linear():
    #     raise NotImplementedError("Only left linear rules are supported for now")

    # store the vertices we have split for id's on the LHS
    in_map: Dict[int, int] = dict()
    out_map: Dict[int, int] = dict()

    # compute the pushout complement
    ctx = m.cod.copy()
    for e in r.lhs.edges():
        ctx.remove_edge(m.emap[e])
    for v in r.lhs.vertices():
        v1 = m.vmap[v]
        if r.lhs.is_boundary(v):
            in_c = len(r.lhs.vertex_data(v).in_indices)
            out_c = len(r.lhs.vertex_data(v).out_indices)
            if in_c == 1 and out_c == 1:
                v1i, v1o = ctx.explode_vertex(v1)
                if len(v1i) == 1 and len(v1o) == 1:
                    in_map[v] = v1i[0]
                    out_map[v] = v1o[0]
                else:
                    raise NotImplementedError("Rewriting modulo Frobenius not yet supported.")
            elif in_c > 1 or out_c > 1:
                raise NotImplementedError("Rewriting modulo Frobenius not yet supported.")
        else:
            ctx.remove_vertex(v1)

    # this will be the rewritten graph
    h = ctx

    # this will embed r.rhs into h
    m1 = Match(r.rhs, h)
    # vmap1: Dict[int,int] = {}

    # first map the inputs, using the matching of the lhs
    for vl,vr in zip(r.lhs.inputs(), r.rhs.inputs()):
        m1.vmap[vr] = in_map[vl] if vl in in_map else m.vmap[vl]

    # next map the outputs. if the same vertex is an input and an output in r.rhs, then
    # merge them in h.
    for vl,vr in zip(r.lhs.outputs(), r.rhs.outputs()):
        vr1 = out_map[vl] if vl in out_map else m.vmap[vl]
        if vr in m1.vmap:
            h.merge_vertices(m1.vmap[vr], vr1)
        else:
            m1.vmap[vr] = vr1

    # then map the interior to new, fresh vertices
    for v in r.rhs.vertices():
        if not r.rhs.is_boundary(v):
            vd = r.rhs.vertex_data(v)
            v1 = h.add_vertex(vd.x, vd.y, vd.value)
            m1.vmap[v] = v1
            m1.vimg.add(v1)

    # now add the edges from rhs to h and connect them using vmap1
    for e in r.rhs.edges():
        ed = r.rhs.edge_data(e)
        e1 = h.add_edge([m1.vmap[v] for v in ed.s],
                        [m1.vmap[v] for v in ed.t],
                        ed.value, ed.x, ed.y, ed.fg, ed.bg, ed.hyper)
        m1.emap[e] = e1
        m1.eimg.add(e1)

    return [m1]

def rewrite(r: Rule, m: Match) -> Graph:
    """Apply the given rewrite r to at match m and return the first result

    This is a convience wrapper for `dpo` for when the extra rewrite data
    isn't needed."""

    try:
        result = next(iter(dpo(r, m)))
        return result.cod
    except StopIteration:
        raise RuntimeError("Rewrite has no valid context")




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import itertools
from typing import Optional

from .graph import Graph

class RuleError(Exception):
    pass

class Rule:
    lhs: Graph
    rhs: Graph

    def __init__(self, lhs: Graph, rhs: Graph, name: str='', equiv: bool=True):
        if len(lhs.inputs()) != len(rhs.inputs()) or len(lhs.outputs()) != len(rhs.outputs()):
            raise RuleError("Inputs and outputs must match on LHS and RHS of rule")
        self.lhs = lhs
        self.rhs = rhs
        self.name = name
        self.equiv = equiv

    def converse(self) -> Optional[Rule]:
        if self.equiv:
            if self.name.startswith('-'):
                name = self.name[1:]
            else:
                name = '-' + self.name

            return Rule(self.rhs.copy(), self.lhs.copy(), name, True)
        else:
            return None

    def is_left_linear(self) -> bool:
        """Returns True if boundary on lhs embeds injectively"""
        verts = set()
        for v in itertools.chain(self.lhs.inputs(), self.lhs.outputs()):
            if v in verts: return False
            verts.add(v)
        return True



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QGraphicsView

from ..graph import Graph
from .graphscene import GraphScene

class GraphView(QGraphicsView):
    def __init__(self) -> None:
        self.graph_scene = GraphScene()
        super().__init__(self.graph_scene)
        self.setMouseTracking(True)
        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorViewCenter)

    def set_graph(self, g: Graph) -> None:
        self.graph_scene.set_graph(g)
        self.centerOn(0,0)



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

from .matcher import find_iso, match_rule
from .parser import parse
from .rewrite import dpo
from .graph import Graph
from .rule import Rule

# class SemanticError(Exception):
#     def __init__(self, line: int, message: str):
#         self.line = line
#         self.message = message
#         super().__init__(str(self.line) + ": " + self.message)


class RewriteState:
    UNCHECKED = 0
    CHECKING = 1
    VALID = 2
    INVALID = 3

    def __init__(self,
                 term_pos: Tuple[int,int] = (0,0),
                 equiv: bool=True,
                 rule: Optional[Rule]=None,
                 lhs: Optional[Graph]=None,
                 rhs: Optional[Graph]=None,
                 lhs_match: Optional[Graph]=None,
                 rhs_match: Optional[Graph]=None,
                 stub: bool=False) -> None:
        self.status = RewriteState.UNCHECKED
        self.term_pos = term_pos
        self.equiv = equiv
        self.rule = rule
        self.lhs = lhs
        self.rhs = rhs
        self.lhs_match = lhs_match
        self.rhs_match = rhs_match
        self.stub = stub

    def check(self) -> None:
        if self.rule and self.lhs and self.rhs:
            # check for any constraints on the LHS and RHS first
            if self.lhs_match and not find_iso(self.lhs, self.lhs_match):
                self.status = RewriteState.INVALID
            if self.rhs_match and not find_iso(self.rhs, self.rhs_match):
                self.status = RewriteState.INVALID

            # if all LHS/RHS constraints are satisfied, try to prove the rule step by rewriting
            if self.status == RewriteState.CHECKING:
                for m_lhs in match_rule(self.rule, self.lhs):
                    for m_rhs in dpo(self.rule, m_lhs):
                        iso = find_iso(m_rhs.cod, self.rhs)
                        if iso:
                            self.status = RewriteState.VALID

                            for v in m_lhs.dom.vertices():
                                self.lhs.vertex_data(m_lhs.vmap[v]).highlight = True
                            for e in m_lhs.dom.edges():
                                self.lhs.edge_data(m_lhs.emap[e]).highlight = True

                            for v in m_rhs.dom.vertices():
                                self.rhs.vertex_data(iso.vmap[m_rhs.vmap[v]]).highlight = True
                            for e in m_rhs.dom.edges():
                                self.rhs.edge_data(iso.emap[m_rhs.emap[e]]).highlight = True

                            break

        if self.status != RewriteState.VALID:
            self.status = RewriteState.INVALID

class State:
    def __init__(self) -> None:
        self.graphs: Dict[str, Graph] = dict()
        self.rules: Dict[str, Rule] = dict()
        self.rewrites: Dict[str, RewriteState] = dict()
        self.parts: List[Tuple[int, int, str, str]] = list()
        self.errors: List[Tuple[str, int, str]] = list()

    def update(self, code: str, file_name: str) -> None:
        parse_data = parse(code, file_name)
        self.graphs = parse_data.graphs
        self.rules = parse_data.rules
        self.parts = parse_data.parts
        self.errors = parse_data.errors

        for name, (t_start, t_end, equiv, rule, lhs, rhs, lhs_match, rhs_match) in parse_data.rewrites.items():
            stub = not (':' in name)
            self.rewrites[name] = RewriteState((t_start, t_end), equiv, rule, lhs, rhs, lhs_match, rhs_match, stub)

    def part_with_index_at(self, pos: int) -> Optional[Tuple[int, Tuple[int,int,str,str]]]:
        p0 = (0, self.parts[0]) if len(self.parts) >= 1 else None
        for (i,p) in enumerate(self.parts):
            if p[0] <= pos:
                p0 = (i,p)
                if p[1] >= pos:
                    return (i,p)
        return p0

    def part_at(self, pos: int) -> Optional[Tuple[int,int,str,str]]:
        p = self.part_with_index_at(pos)
        return p[1] if p else None




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Tuple, Union
from PySide6.QtCore import QFileInfo, QPersistentModelIndex, Qt, QAbstractItemModel, QModelIndex
from PySide6.QtGui import QFont

class ErrorListModel(QAbstractItemModel):
    """A model containing a list of errors, with line numbers"""
    errors: List[Tuple[str,int,str]]

    def __init__(self) -> None:
        super().__init__()
        self.errors = []

    def set_errors(self, errors: List[Tuple[str, int, str]]) -> None:
        self.beginResetModel()
        self.errors = errors
        self.endResetModel()

    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: int=Qt.ItemDataRole.DisplayRole) -> Any:
        """Overrides `QAbstractItemModel.data` to populate a view with errors"""

        if index.row() >= len(self.errors) or index.column() >= 3:
            return None

        if role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return QFileInfo(self.errors[index.row()][0]).fileName()
            else:
                return self.errors[index.row()][index.column()] 
        elif role == Qt.ItemDataRole.FontRole:
            return QFont("monospace", 12)
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return Qt.AlignmentFlag.AlignTop

    def headerData(self, section: int, orientation: Qt.Orientation, role: int=Qt.ItemDataRole.DisplayRole) -> Any:
        """Overrides `QAbstractItemModel.headerData` to populate a view with column names"""

        if role == Qt.ItemDataRole.DisplayRole:
            if section == 0: return "File"
            elif section == 1: return "Line"
            else: return "Error"
        else:
            return None

    def index(self, row: int, column: int, parent: Union[QModelIndex, QPersistentModelIndex]=QModelIndex()) -> QModelIndex:
        """Construct a `QModelIndex` for the given row and column"""

        if not self.hasIndex(row, column, parent): return QModelIndex()
        else: return self.createIndex(row, column, None)

    def columnCount(self, index: Union[QModelIndex, QPersistentModelIndex]=QModelIndex()) -> int:
        """The number of columns"""
        return 3

    def rowCount(self, index: Union[QModelIndex, QPersistentModelIndex]=QModelIndex()) -> int:
        """The number of rows"""
        if not index or not index.isValid(): return len(self.errors)
        else: return 0

    def parent(self, child: Any=None) -> Any:
        """Always return an invalid index, since there are no nested indices"""
        if child is None:
            return super().parent()
        else:
            return QModelIndex()




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Iterable, Set, List, Dict, Iterator, Any, Optional, Tuple
import json
import copy

class GraphError(Exception):
    pass

class VData:
    def __init__(self, x: float=0, y: float=0, value: Any="") -> None:
        self.value = value
        self.x = x
        self.y = y
        self.highlight = False

        # for quickly finding edges in the nhd of v
        self.in_edges: Set[int] = set()
        self.out_edges: Set[int] = set()

        # if v is in the list of inputs/outputs, save its index here
        self.in_indices: Set[int] = set()
        self.out_indices: Set[int] = set()

class EData:
    def __init__(self,
            s: Optional[List[int]]=None,
            t: Optional[List[int]]=None,
            value: Any="",
            x: float=0,
            y: float=0,
            fg: str='',
            bg: str='',
            hyper: bool=True) -> None:
        self.value = value
        self.highlight = False
        self.x = x
        self.y = y
        self.s = [] if s is None else s
        self.t = [] if t is None else t
        self.fg = fg
        self.bg = bg
        self.hyper = hyper

    def __repr__(self) -> str:
        return "Edge: %s (%d, %d)" % (self.value, self.x, self.y)

    def box_size(self) -> int:
        """Returns the number of 'units' of width the box should have to display nicely.

        The simple rule is if both inputs and outputs are <= 1, draw as a small (size 1) box, otherwise
        draw as a larger (size 2) box."""
        return 1 if len(self.s) <= 1 and len(self.t) <= 1 else 2

class Graph:
    def __init__(self) -> None:
        self.vdata: Dict[int, VData] = {}
        self.edata: Dict[int, EData] = {}
        self._inputs: List[int] = []
        self._outputs: List[int] = []
        self.vindex = 0
        self.eindex = 0

    def copy(self) -> Graph:
        g = Graph()
        g.vdata = copy.deepcopy(self.vdata)
        g.edata = copy.deepcopy(self.edata)
        g._inputs = self._inputs.copy()
        g._outputs = self._outputs.copy()
        g.vindex = self.vindex
        g.eindex = self.eindex
        return g

    def vertices(self) -> Iterator[int]:
        return iter(self.vdata.keys())

    def edges(self) -> Iterator[int]:
        return iter(self.edata.keys())

    def num_vertices(self) -> int:
        return len(self.vdata)

    def num_edges(self) -> int:
        return len(self.edata)

    def vertex_data(self, v: int) -> VData:
        return self.vdata[v]

    def edge_data(self, e: int) -> EData:
        return self.edata[e]

    def in_edges(self, v: int) -> Set[int]:
        return self.vdata[v].in_edges

    def out_edges(self, v: int) -> Set[int]:
        return self.vdata[v].out_edges

    def source(self, e: int) -> List[int]:
        return self.edata[e].s

    def target(self, e: int) -> List[int]:
        return self.edata[e].t

    def add_vertex(self, x:float=0, y:float=0, value: Any="", name: int=-1) -> int:
        """Add a vertex to the graph
        
        :param x:     The X coordinate to draw the vertex
        :param y:     The Y coordinate
        :param value: The value carried by this vertex. This is the empty string by default.
        :param name:  An optional name. If this is set to -1, set the name automatically. After
                      setting names manually, be sure to call `update_indices` to make sure
                      future automatic names don't clash.
        """
        if name == -1:
            v = self.vindex
            self.vindex += 1
        else:
            v = name

        self.vdata[v] = VData(x, y, value)
        return v

    def add_edge(self, s:List[int], t:List[int], value:Any="", x:float=0, y:float=0, fg:str='', bg:str='', hyper:bool=True, name:int=-1) -> int:
        """Add an edge to the graph
        
        :param s:     A list of source vertices
        :param t:     A list of target vertices
        :param value: The value carried by this edge (typically a string)
        :param x:     The X coordinate to draw the box representing this hyperedge
        :param y:     The Y coordinate
        :param hyper: This is a hint to tell the GUI how to draw this (hyper)edge. If set to
                      False, ideally it should be drawn simply as a line connected two vertices
                      rather than as a box.
        :param name:  An optional name. If this is set to -1, set the name automatically. After
                      setting names manually, be sure to call `update_indices` to make sure
                      future automatic names don't clash.
        """
        if name == -1:
            e = self.eindex
            self.eindex += 1
        else:
            e = name

        self.edata[e] = EData(s, t, value, x, y, fg, bg, hyper)
        for v in s: self.vdata[v].out_edges.add(e)
        for v in t: self.vdata[v].in_edges.add(e)
        return e

    def remove_vertex(self, v: int, strict: bool=False) -> None:
        if strict:
            if (len(self.vertex_data(v).in_edges)  > 0 or
                len(self.vertex_data(v).out_edges) > 0):
                raise Exception("Attempting to remove vertex with adjacent edges")
            if (v in self.inputs() or v in self.outputs()):
                raise Exception("Attempting to remove boundary vertex")
        else:
            for e in self.vertex_data(v).in_edges:
                self.edge_data(e).t = [v1 for v1 in self.edge_data(e).t if v1 != v]
            for e in self.vertex_data(v).out_edges:
                self.edge_data(e).s = [v1 for v1 in self.edge_data(e).s if v1 != v]
            self.set_inputs([v1 for v1 in self.inputs() if v1 != v])
            self.set_outputs([v1 for v1 in self.outputs() if v1 != v])
        del self.vdata[v]

    def remove_edge(self, e: int) -> None:
        for v in self.edge_data(e).s:
            self.vertex_data(v).out_edges.remove(e)
        for v in self.edge_data(e).t:
            self.vertex_data(v).in_edges.remove(e)
        del self.edata[e]

    def add_simple_edge(self, s:int, t:int, value: Any="") -> int:
        e = self.add_edge([s], [t], value, hyper=False)
        return e

    def update_indices(self) -> None:
        self.vindex = max(self.vdata.keys()) + 1 if self.vdata else 0
        self.eindex = max(self.edata.keys()) + 1 if self.edata else 0

    def set_inputs(self, inp: List[int]) -> None:
        self._inputs = inp
        for d in self.vdata.values():
            d.in_indices.clear()

        for i,v in enumerate(self._inputs):
            self.vdata[v].in_indices.add(i)

    def set_outputs(self, outp: List[int]) -> None:
        self._outputs = outp
        for d in self.vdata.values():
            d.out_indices.clear()

        for i,v in enumerate(self._outputs):
            self.vdata[v].out_indices.add(i)

    def inputs(self) -> List[int]:
        return self._inputs

    def outputs(self) -> List[int]:
        return self._outputs

    def is_input(self, v: int) -> bool:
        return len(self.vdata[v].in_indices) > 0

    def is_output(self, v: int) -> bool:
        return len(self.vdata[v].out_indices) > 0

    def is_boundary(self, v: int) -> bool:
        return self.is_input(v) or self.is_output(v)

    def successors(self, vs: Iterable[int]) -> Set[int]:
        succ: Set[int] = set()
        current = list(vs)
        while len(current) > 0:
            v = current.pop()
            for e in self.out_edges(v):
                for v1 in self.target(e):
                    if v1 not in succ:
                        succ.add(v1)
                        current.append(v1)

        return succ

    def merge_vertices(self, v: int, w: int) -> None:
        """Identify the two vertices given

        Form the quotient of the graph by identifying v with w. Afterwards, the
        quotiented vertex will be named v.
        """

        vd = self.vertex_data(v)
        
        # print("merging %s <- %s" % (v, w))
        for e in self.in_edges(w):
            ed = self.edge_data(e)
            ed.t = [v if x == w else x for x in ed.t]
            vd.in_edges.add(e)

        for e in self.out_edges(w):
            ed = self.edge_data(e)
            ed.s = [v if x == w else x for x in ed.s]
            vd.out_edges.add(e)

        self.set_inputs([v if x == w else x for x in self.inputs()])
        self.set_outputs([v if x == w else x for x in self.outputs()])
        self.remove_vertex(w)

    def explode_vertex(self, v: int) -> Tuple[List[int], List[int]]:
        """Split a vertex into one copy for each input, in-tentacle, output, and out-tentacle

        This is used for computing pushout complements of rules that aren't left-linear. Returns
        a pair of lists containing the new input-like and output-like vertices, respectively.
        """

        new_vs: Tuple[List[int], List[int]] = ([], [])
        vd = self.vertex_data(v)
        def fresh(j: int) -> int:
            v1 = self.add_vertex(vd.x, vd.y, vd.value)
            new_vs[j].append(v1)
            return v1

        self.set_inputs([v1 if v1 != v else fresh(0) for v1 in self.inputs()])

        for e in vd.in_edges:
            ed = self.edge_data(e)
            for i in range(len(ed.t)):
                if ed.t[i] == v:
                    ed.t[i] = fresh(0)
                    self.vertex_data(ed.t[i]).in_edges.add(e)

        self.set_outputs([v1 if v1 != v else fresh(1) for v1 in self.outputs()])

        for e in vd.out_edges:
            ed = self.edge_data(e)
            for i in range(len(ed.s)):
                if ed.s[i] == v:
                    ed.s[i] = fresh(1)
                    self.vertex_data(ed.s[i]).out_edges.add(e)


        vd.in_edges = set()
        vd.out_edges = set()
        self.remove_vertex(v, strict=True)

        return new_vs

    def insert_id_after(self, v: int, reverse: bool = False) -> int:
        """Insert a new identity hyperedge after the given vertex

        Insert a dummy identity box with source at the given vertex and redirect
        any out-edges or outputs to the target of the new hyperedge. If `reverse`
        is True, then flip the source and target of the identity wire. This can be
        used to break directed cycles, essentially by introducing a cap and cup.
        """
        vd = self.vertex_data(v)
        w = self.add_vertex(vd.x + 3, vd.y, vd.value)
        wd = self.vertex_data(w)
        wd.highlight = vd.highlight
        self.set_outputs([x if x != v else w for x in self.outputs()])
        for e in vd.out_edges:
            ed = self.edge_data(e)
            ed.s = [x if x != v else w for x in ed.s]
            wd.out_edges.add(e)
        vd.out_edges.clear()

        s, t = ([v], [w]) if not reverse else ([w], [v])
        e = self.add_edge(s, t, "id", vd.x + 1.5, vd.y)
        self.edge_data(e).highlight = vd.highlight
        return e
    
    def tensor(self, other: Graph) -> None:
        """Take the monoidal product with the given graph

        Calling g.tensor(h) will turn g into g ⊗ h. Use the infix version "g + h" to simply return
        the tensor product without changing g.
        """
        vmap = dict()
        # emap = dict()

        max_self = max(max((self.vertex_data(v).y for v in self.vertices()), default = 0),
                       max((self.edge_data(e).y for e in self.edges()), default=0))
        min_other = min(min((other.vertex_data(v).y for v in other.vertices()), default = 0),
                        min((other.edge_data(e).y for e in other.edges()), default=0))

        for v in self.vertices(): self.vertex_data(v).y -= max_self
        for e in self.edges(): self.edge_data(e).y -= max_self

        for v in other.vertices():
            vd = other.vertex_data(v)
            vmap[v] = self.add_vertex(vd.x, vd.y - min_other + 1, vd.value)

        for e in other.edges():
            ed = other.edge_data(e)
            self.add_edge([vmap[v] for v in ed.s],
                          [vmap[v] for v in ed.t],
                          ed.value, ed.x, ed.y - min_other + 1, ed.fg, ed.bg, ed.hyper)
        
        self.set_inputs(self.inputs() + [vmap[v] for v in other.inputs()])
        self.set_outputs(self.outputs() + [vmap[v] for v in other.outputs()])

    def __mul__(self, other: Graph) -> Graph:
        g = self.copy()
        g.tensor(other)
        return g

    def compose(self, other: Graph) -> None:
        """Compose with given graph in diagram order"""

        vmap = dict()

        max_self = max(max((self.vertex_data(v).x for v in self.vertices()), default = 0),
                       max((self.edge_data(e).x for e in self.edges()), default=0))
        min_other = min(min((other.vertex_data(v).x for v in other.vertices()), default = 0),
                        min((other.edge_data(e).x for e in other.edges()), default=0))

        for v in self.vertices(): self.vertex_data(v).x -= max_self
        for e in self.edges(): self.edge_data(e).x -= max_self

        for v in other.vertices():
            vd = other.vertex_data(v)
            vmap[v] = self.add_vertex(vd.x - min_other, vd.y, vd.value)

        for e in other.edges():
            ed = other.edge_data(e)
            self.add_edge([vmap[v] for v in ed.s],
                          [vmap[v] for v in ed.t],
                          ed.value, ed.x - min_other, ed.y, ed.fg, ed.bg, ed.hyper)
        
        plug1 = self.outputs()
        plug2 = [vmap[v] for v in other.inputs()]
        quotient: Dict[int,int] = dict()
        if len(plug1) != len(plug2):
            raise GraphError("Attempting to plug a graph with %d outputs into one with %d inputs" % (len(plug1), len(plug2)))

        self.set_outputs([vmap[v] for v in other.outputs()])

        for i in range(len(plug1)):
            p1 = plug1[i]
            p2 = plug2[i]
            while p1 in quotient: p1 = quotient[p1]
            while p2 in quotient: p2 = quotient[p2]
            if p1 != p2:
                self.merge_vertices(p1, p2)
                quotient[p2] = p1

    def __rshift__(self, other: Graph) -> Graph:
        g = self.copy()
        g.compose(other)
        return g

def gen(value: str, arity: int, coarity: int, fg: str='', bg: str='') -> Graph:
    g = Graph()
    inputs = [g.add_vertex(-1.5, i - (arity-1)/2) for i in range(arity)]
    outputs = [g.add_vertex(1.5, i - (coarity-1)/2) for i in range(coarity)]
    g.add_edge(inputs, outputs, value, fg=fg, bg=bg)
    g.set_inputs(inputs)
    g.set_outputs(outputs)
    return g
        
def perm(p: List[int]) -> Graph:
    g = Graph()
    size = len(p)
    inputs = [g.add_vertex(0, i - (size-1)/2) for i in range(size)]
    outputs = [inputs[p[i]] for i in range(size)]
    g.set_inputs(inputs)
    g.set_outputs(outputs)
    return g

def identity() -> Graph:
    g = Graph()
    v = g.add_vertex(0, 0)
    g.set_inputs([v])
    g.set_outputs([v])
    return g

# def wide_id() -> Graph:
#     return gen("id", 1, 1)

def id_perm(p: List[int]) -> Graph:
    g = Graph()
    size = len(p)
    inputs = [g.add_vertex(-1.5, i - (size-1)/2) for i in range(size)]
    outputs = [g.add_vertex(1.5, i - (size-1)/2) for i in range(size)]

    for i in range(size):
        y = i - (size-1)/2
        g.add_edge([inputs[i]], [outputs[p[i]]], "id", 0, y)

    g.set_inputs(inputs)
    g.set_outputs(outputs)

    return g

def load_graph(path: str) -> Graph:
    """Load a .chyp graph file from the given path"""

    with open(path) as f:
        g = graph_from_json(f.read())
    return g

def graph_from_json(json_string: str) -> Graph:
    """Load a graph from the given JSON string"""

    j = json.loads(json_string)
    g = Graph()
    for v,vd in j["vertices"].items():
        g.add_vertex(x=float(vd["x"] if "x" in vd else 0.0),
                     y=float(vd["y"] if "y" in vd else 0.0),
                     value=vd["value"] if "value" in vd else "",
                     name=int(v))
    for e,ed in j["edges"].items():
        g.add_edge(s=[int(v) for v in ed["s"]],
                   t=[int(v) for v in ed["t"]],
                   value=ed["value"] if "value" in ed else "",
                   x=float(ed["x"]) if "x" in ed else 0.0,
                   y=float(ed["y"]) if "y" in ed else 0.0,
                   hyper=bool(ed["hyper"]) if "hyper" in ed else True,
                   name=int(e))

    g.set_inputs([int(v) for v in j["inputs"]])
    g.set_outputs([int(v) for v in j["outputs"]])
    g.update_indices()
    return g



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Tuple
import re
from PySide6.QtGui import QColor, QSyntaxHighlighter, QTextDocument

from .colors import current_theme

NO_STATUS = 0
STATUS_GOOD = 1
STATUS_BAD = 2

# palette from: https://github.com/catppuccin/catppuccin
theme = current_theme()
FG =      theme['fg']
BG =      theme['bg']
SEL =     theme['bg_sel']
KEYWORD = theme['fg_keyword']
IDENT =   theme['fg_ident']
NUM =     theme['fg_num']
OP =      theme['fg_op']
COMMENT = theme['fg_comment']
STRING  = theme['fg_string']
BG_GOOD = theme['bg_good']
BG_BAD  = theme['bg_bad']

class ChypHighlighter(QSyntaxHighlighter):
    def __init__(self, doc: QTextDocument) -> None:
        super().__init__(doc)
        self.region: Optional[Tuple[int,int]] = None
        self.region_status = NO_STATUS

    def set_current_region(self, region: Optional[Tuple[int,int]], status: int) -> None:
        self.region = region
        self.region_status = status
        self.rehighlight()

    def highlightBlock(self, text: str) -> None:
        ident = '[a-zA-Z_][\\.a-zA-Z0-9_]*'
        for m in re.finditer('(^|\\W)(let|def|gen|rule|by|rewrite|import|as|show)\\s+\\-?\\s*(%s)?' % ident, text):
            x,y = m.span(2)
            self.setFormat(x, y-x, QColor(KEYWORD))
            x,y = m.span(3)
            self.setFormat(x, y-x, QColor(IDENT))

        for m in re.finditer('[?~.><:;*=\\-\\[\\]]', text):
            self.setFormat(m.start(), 1, QColor(OP))

        for m in re.finditer('(\\W|^)([0-9]+)(\\W|$)', text):
            x,y = m.span(2)
            self.setFormat(x, y-x, QColor(NUM))

        for m in re.finditer('#.*$', text):
            x,y = m.span(0)
            self.setFormat(x, y-x, QColor(COMMENT))

        for m in re.finditer('"[^"]*"', text):
            x,y = m.span(0)
            self.setFormat(x, y-x, QColor(STRING))

        # highlight the region that is currently in focus
        if self.region:
            x, y = self.region
            start = self.currentBlock().position()
            length = self.currentBlock().length()
            end = start + length
            if y >= start and x < end:
                x_rel = max(x - start, 0)
                y_rel = min(y - start, length)
                for c in range(x_rel, y_rel):
                    f = self.format(c)

                    if self.region_status == STATUS_GOOD:
                        f.setBackground(QColor(BG_GOOD))
                    elif self.region_status == STATUS_BAD:
                        f.setBackground(QColor(BG_BAD))
                    else:
                        f.setBackground(QColor(SEL))
                    self.setFormat(c, 1, f)




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Callable, List, Optional
from PySide6.QtCore import QByteArray, QDir, QFileInfo, QSettings, Qt
from PySide6.QtGui import QActionGroup, QCloseEvent, QKeySequence
from PySide6.QtWidgets import QApplication, QFileDialog, QMainWindow, QMenuBar, QMessageBox, QTabWidget, QVBoxLayout, QWidget

from .editor import Editor


class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        conf = QSettings('chyp', 'chyp')

        self.setWindowTitle("chyp")

        self.tabs = QTabWidget()
        self.tabs.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.tabs.currentChanged.connect(self.tab_changed)

        w = QWidget(self)
        w.setLayout(QVBoxLayout())
        self.setCentralWidget(w)
        w.layout().setContentsMargins(0,0,0,0)
        w.layout().setSpacing(0)
        w.layout().addWidget(self.tabs)
        self.resize(1600, 800)
        
        geom = conf.value("editor_window_geometry")
        if geom and isinstance(geom, QByteArray): self.restoreGeometry(geom)
        self.show()

        self.active_editor: Optional[Editor] = None
        self.add_tab(Editor(), "Untitled")
        self.update_file_name()
        self.build_menu()


    def remove_empty_editor(self) -> None:
        if self.active_editor:
            if self.active_editor.title() == 'Untitled' and self.active_editor.doc.toPlainText() == '':
                self.tabs.removeTab(self.tabs.indexOf(self.active_editor))
                self.active_editor = None

    def update_file_name(self) -> None:
        for i in range(self.tabs.count()):
            w = self.tabs.widget(i)
            if isinstance(w, Editor):
                self.tabs.setTabText(i, w.title())
        if self.active_editor:
            self.setWindowTitle('chyp - ' + self.active_editor.title())

    def tab_changed(self, i: int) -> None:
        w = self.tabs.widget(i)
        if isinstance(w, Editor):
            self.active_editor = w
            self.active_editor.code_view.setFocus()
            self.update_file_name()

    def update_themes(self) -> None:
        conf = QSettings('chyp', 'chyp')
        theme_name = conf.value('theme')
        if not theme_name or not isinstance(theme_name, str):
            theme_name = 'catppuccin_macchiato'

        def set_th(t: str) -> Callable:
            def f() -> None:
                conf.setValue('theme', t)
                QMessageBox.information(self, 'Theme set',
                                        'You must restart Chyp for the new theme to take effect.')
            return f

        themes_group = QActionGroup(self)

        view_themes_dark = self.view_themes.addAction("Dark")
        view_themes_dark.setCheckable(True)
        view_themes_dark.setChecked(theme_name == 'catppuccin_macchiato')
        view_themes_dark.triggered.connect(set_th('catppuccin_macchiato'))

        view_themes_light = self.view_themes.addAction("Light")
        view_themes_light.setCheckable(True)
        view_themes_light.setChecked(theme_name == 'catppuccin_latte')
        view_themes_light.triggered.connect(set_th('catppuccin_latte'))

        themes_group.addAction(view_themes_dark)
        themes_group.addAction(view_themes_light)

    def recent_files(self) -> List[str]:
        conf = QSettings('chyp', 'chyp')
        o = conf.value('recent_files', [])
        return o if isinstance(o, list) else []

    def update_recent_files(self) -> None:
        def open_recent(f: str) -> Callable:
            return lambda: self.open(f)

        self.file_open_recent.clear()
        for f in self.recent_files():
            fi = QFileInfo(f)
            action = self.file_open_recent.addAction(fi.fileName())
            action.triggered.connect(open_recent(f))

    def add_tab(self, editor: Editor, title: str) -> None:
        self.tabs.addTab(editor, title)
        editor.doc.fileNameChanged.connect(self.update_file_name)
        editor.doc.modificationChanged.connect(self.update_file_name)
        self.tabs.setCurrentWidget(editor)
        editor.reset_state()

    def close_tab(self, editor: Optional[Editor]=None) -> bool:
        if editor is None:
            editor = self.active_editor

        if editor:
            i = self.tabs.indexOf(editor)
            if i != -1 and editor.doc.confirm_close():
                editor.doc.fileNameChanged.disconnect(self.update_file_name)
                editor.doc.modificationChanged.disconnect(self.update_file_name)
                self.tabs.removeTab(i)

                if self.tabs.count() == 0:
                    app = QApplication.instance()
                    if app:
                        app.quit()

                return True

        return False

    def new(self) -> None:
        self.remove_empty_editor()
        editor = Editor()
        self.add_tab(editor, "Untitled")

    def open(self, file_name: str='') -> None:
        conf = QSettings('chyp', 'chyp')

        # if no file name provided, show open dialog
        if file_name == '':
            o = conf.value('last_dir')
            last_dir = o if isinstance(o, str) else QDir.home().absolutePath()
            file_name, _ = QFileDialog.getOpenFileName(self,
                                                       "Open File",
                                                       last_dir,
                                                       'chyp files (*.chyp)')

        # if file is already open, just focus the tab
        for i in range(self.tabs.count()):
            w = self.tabs.widget(i)
            if isinstance(w, Editor) and w.doc.file_name == file_name:
                self.tabs.setCurrentWidget(w)
                return

        try:
            editor = Editor()
            editor.doc.open(file_name)
            conf.setValue('last_dir', QFileInfo(file_name).absolutePath())
            self.remove_empty_editor()
            self.update_recent_files()
            editor.doc.fileNameChanged.connect(self.update_file_name)
            self.add_tab(editor, editor.title())
        except FileNotFoundError:
            QMessageBox.warning(self, "File not found", "File not found: " + file_name)


    def save(self) -> None:
        if self.active_editor:
            self.active_editor.doc.save()
            self.update_recent_files()

    def save_as(self) -> None:
        if self.active_editor:
            self.active_editor.doc.save_as()
            self.update_recent_files()
    
    def undo(self) -> None:
        if self.active_editor:
            self.active_editor.code_view.undo()

    def redo(self) -> None:
        if self.active_editor:
            self.active_editor.code_view.redo()

    def show_errors(self) -> None:
        if self.active_editor:
            self.active_editor.show_errors()

    def add_rewrite_step(self) -> None:
        if self.active_editor:
            self.active_editor.code_view.add_line_below("  = ? by ")

    def repeat_rewrite_step(self) -> None:
        if self.active_editor:
            self.active_editor.repeat_step_at_cursor()

    def next_rewrite(self) -> None:
        if self.active_editor:
            self.active_editor.next_rewrite_at_cursor()

    def next_part(self) -> None:
        if self.active_editor:
            self.active_editor.next_part(step=1)

    def previous_part(self) -> None:
        if self.active_editor:
            self.active_editor.next_part(step=-1)

    def next_tab(self) -> None:
        c = self.tabs.count()
        if c != 0:
            i = (self.tabs.currentIndex() + 1) % c
            self.tabs.setCurrentIndex(i)

    def previous_tab(self) -> None:
        c = self.tabs.count()
        if c != 0:
            i = (self.tabs.currentIndex() - 1) % c
            self.tabs.setCurrentIndex(i)

    def goto_import(self) -> None:
        if self.active_editor:
            f = self.active_editor.import_at_cursor()
            if f:
                self.open(f)


    def closeEvent(self, e: QCloseEvent) -> None:
        conf = QSettings('chyp', 'chyp')
        conf.setValue("editor_window_geometry", self.saveGeometry())

        if self.active_editor:
            conf.setValue("editor_splitter_state", self.active_editor.splitter.saveState())
            sizes = self.active_editor.splitter.sizes()
            if sizes[2] != 0:
                conf.setValue('error_panel_size', sizes[2])

        while self.tabs.count() > 0:
            w = self.tabs.widget(0)
            if isinstance(w, Editor):
                if not self.close_tab(w):
                    e.ignore()
                    return
            else:
                raise RuntimeError("Unexpected widget in tab list")

        e.accept()

    def build_menu(self) -> None:
        menu = QMenuBar()
        file_menu = menu.addMenu("&File")
        edit_menu = menu.addMenu("&Edit")
        code_menu = menu.addMenu("&Code")
        view_menu = menu.addMenu("&View")

        file_new = file_menu.addAction("&New")
        file_new.triggered.connect(lambda: self.new())

        file_open = file_menu.addAction("&Open")
        file_open.setShortcut(QKeySequence(QKeySequence.StandardKey.Open))
        file_open.triggered.connect(lambda: self.open())

        self.file_open_recent = file_menu.addMenu("Open &Recent")
        self.update_recent_files()

        file_menu.addSeparator()

        file_close = file_menu.addAction("&Close")
        file_close.setShortcut(QKeySequence(QKeySequence.StandardKey.Close))
        file_close.triggered.connect(lambda: self.close_tab())

        file_save = file_menu.addAction("&Save")
        file_save.setShortcut(QKeySequence(QKeySequence.StandardKey.Save))
        file_save.triggered.connect(lambda: self.save())

        file_save_as = file_menu.addAction("Save &As")
        file_save_as.setShortcut(QKeySequence(QKeySequence.StandardKey.SaveAs))
        file_save_as.triggered.connect(lambda: self.save_as())

        file_menu.addSeparator()

        file_exit = file_menu.addAction("E&xit")
        file_exit.setShortcut(QKeySequence(QKeySequence.StandardKey.Quit))

        app = QApplication.instance()
        if app:
            file_exit.triggered.connect(app.quit)

        edit_undo = edit_menu.addAction("&Undo")
        edit_undo.setShortcut(QKeySequence(QKeySequence.StandardKey.Undo))
        edit_undo.triggered.connect(lambda: self.undo())

        edit_redo = edit_menu.addAction("&Redo")
        edit_redo.setShortcut(QKeySequence(QKeySequence.StandardKey.Redo))
        edit_redo.triggered.connect(lambda: self.redo())

        # code_run = code_menu.addAction("&Run")
        # code_run.setShortcut(QKeySequence("Ctrl+R"))
        # code_run.triggered.connect(self.update_state)

        code_show_errors = code_menu.addAction("Show &Errors")
        code_show_errors.setShortcut(QKeySequence("F4"))
        code_show_errors.triggered.connect(lambda: self.show_errors())

        code_add_rewrite_step = code_menu.addAction("&Add Rewrite Step")
        code_add_rewrite_step.setShortcut(QKeySequence("Ctrl+Return"))
        code_add_rewrite_step.triggered.connect(lambda: self.add_rewrite_step())

        code_add_rewrite_step = code_menu.addAction("&Repeat Rewrite Step")
        code_add_rewrite_step.setShortcut(QKeySequence("Ctrl+Shift+Return"))
        code_add_rewrite_step.triggered.connect(lambda: self.repeat_rewrite_step())

        code_next_rewrite = code_menu.addAction("&Next Rewrite")
        code_next_rewrite.setShortcut(QKeySequence("Ctrl+N"))
        code_next_rewrite.triggered.connect(lambda: self.next_rewrite())

        code_menu.addSeparator()

        code_next_part = code_menu.addAction("Next &Part")
        code_next_part.setShortcut(QKeySequence("Ctrl+J"))
        code_next_part.triggered.connect(lambda: self.next_part())

        code_previous_part = code_menu.addAction("Previous &Part")
        code_previous_part.setShortcut(QKeySequence("Ctrl+K"))
        code_previous_part.triggered.connect(lambda: self.previous_part())

        view_next_tab = view_menu.addAction("&Next Tab")
        view_next_tab.setShortcut(QKeySequence("Ctrl+]"))
        view_next_tab.triggered.connect(lambda: self.next_tab())

        view_previous_tab = view_menu.addAction("&Previous Tab")
        view_previous_tab.setShortcut(QKeySequence("Ctrl+["))
        view_previous_tab.triggered.connect(lambda: self.previous_tab())

        view_menu.addSeparator()
        view_goto_import = view_menu.addAction("&Go To Import")
        view_goto_import.setShortcut(QKeySequence("Ctrl+G"))
        view_goto_import.triggered.connect(lambda: self.goto_import())

        view_menu.addSeparator()
        self.view_themes = view_menu.addMenu("&Themes")
        self.update_themes()

        self.setMenuBar(menu)



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import math
import cvxpy as cp
from cvxpy.expressions.variable import Variable
from cvxpy.expressions.constants.constant import Constant
from cvxpy.problems.objective import Minimize
from cvxpy.problems.problem import Problem

from .graph import Graph
from .term import layer_decomp

def convex_layout(g: Graph) -> None:
    """A layout based on `layer_decomp` and convex optimisation

    Vertices and edges are placed in layers according to `layer_decomp`. Their
    y-coordinates are chosen by convex optimation to try to make connections as
    straight as possible subject to the constraints:
      1. inputs and outputs must be in order and at least 1.0 apart
      2. edges must be in order and not overlapping
    """
    e_layers = layer_decomp(g)

    # initialise x-coordinates and rough y-coordinates
    x = -(len(e_layers) + 1) * 1.5
    inp = list(g.inputs())
    for i, v in enumerate(inp):
        vd = g.vertex_data(v)
        vd.x = x + 1.5
        vd.y = i - (len(inp)-1)/2

    x += 3.0

    for e_layer in e_layers:
        v_layer = []
        for i, e in enumerate(e_layer):
            ed = g.edge_data(e)
            ed.x = x
            ed.y = 2 * i - (len(e_layer)-1)
            v_layer += ed.t

        for i, v in enumerate(v_layer):
            vd = g.vertex_data(v)
            vd.x = x + 0.7
            vd.y = i - (len(v_layer)-1)/2
        x += 3.0

    outp = list(g.outputs())
    for i, v in enumerate(outp):
        vd = g.vertex_data(v)
        vd.x = x - 1.5
        vd.y = i - (len(outp)-1)/2

    if g.num_vertices() == 0 or g.num_edges() == 0: return

    # solve for better y-coordinates using convex optimisation
    
    # variables for the y-coordinates of vertices/edges
    vy = Variable(g.num_vertices(), 'vy')
    ey = Variable(g.num_edges(), 'ey')

    # maintain a table from vertex/edge names to variable indices
    vtab = { v : i for i, v in enumerate(g.vertices()) }
    etab = { e : i for i, e in enumerate(g.edges()) }

    constr = []
    opt = []
    # opt = [Constant(0.1) * vy[i] for i in range(g.num_edges())]

    for vlist in (g.inputs(), g.outputs()):
        for i in range(len(vlist) - 1):
            v1 = vlist[i]
            v2 = vlist[i+1]
            constr.append(vy[vtab[v2]] - vy[vtab[v1]] >= Constant(1))
            opt.append(Constant(0.1) * (vy[vtab[v2]] - vy[vtab[v1]]))

    for e_layer in e_layers:
        for i in range(len(e_layer)):
            e1 = e_layer[i]
            if i+1 >= len(e_layer): break
            e2 = e_layer[i+1]
            dist = (g.edge_data(e1).box_size() + g.edge_data(e2).box_size()) * 0.5
            constr.append(ey[etab[e2]] - ey[etab[e1]] >= Constant(dist))
            opt.append(Constant(0.1) * (ey[etab[e2]] - ey[etab[e1]]))

    for v in g.vertices():
        pos1 = vy[vtab[v]]
        pos2 = vy[vtab[v]]
        if len(g.in_edges(v)) >= 1:
            e = next(iter(g.in_edges(v)))
            t = g.target(e)
            y_shift = Constant(0.0 if len(t) <= 1 else ((t.index(v) / (len(t) - 1)) - 0.5))
            pos1 = ey[etab[e]] + y_shift

        if len(g.out_edges(v)) >= 1:
            e = next(iter(g.out_edges(v)))
            s = g.source(e)
            y_shift = Constant(0.0 if len(s) <= 1 else ((s.index(v) / (len(s) - 1)) - 0.5))
            pos2 = ey[etab[e]] + y_shift

        opt.append(pos1 - pos2)


    # problem = Problem(Minimize(cp.sum_squares(cp.vstack(opt))), constr)
    problem = Problem(Minimize(cp.norm1(cp.vstack(opt))), constr)
    problem.solve()
    min = None
    max = None
    for v,i in vtab.items():
        if not vy.value is None:
            y = vy.value[i]
            if min is None or y < min: min = y
            if max is None or y > max: max = y
            g.vertex_data(v).y = y

    if not min is None and not max is None:
        yshift = (min + max) * 0.5
        for v in g.vertices():
            g.vertex_data(v).y -= yshift
    else:
        yshift = 0

    for e,i in etab.items():
        y = ey.value[i]
        ed = g.edge_data(e)
        ed.y = y - yshift
        for j,v in enumerate(ed.t):
            if not g.is_boundary(v):
                yshift_v = 0 if len(ed.t) <= 1 else ((j / (len(ed.t) - 1)) - 0.5)
                g.vertex_data(v).y = ed.y + yshift_v


# def layer_layout(g: Graph) -> None:
#     """A simple layout using `layer_decomp`.

#     Vertices are evenly spaced around the x-axis and edges are placed
#     as close to the average y-coordinate of their inputs as possible.
#     """
#     v_layers, e_layers = layer_decomp(g)
#     x = -(len(v_layers) - 1) * 1.5
#     for l in range(len(e_layers)):
#         v_layer = v_layers[l]
#         e_layer = e_layers[l]
#         for i, v in enumerate(v_layer):
#             vd = g.vertex_data(v)
#             vd.x = x
#             vd.y = i - ((len(v_layer) - 1) * 0.5)

#         # place edges, starting from the middle and working outward
#         end = len(e_layer)
#         start = math.ceil(end/2)

#         max_y = 0.0 # max y-coordinate for edges above the middle
#         min_y = 0.0 # min y-coordinate for edges below the middle

#         # for an odd number of edges, place the middle edge first
#         if start > end/2:
#             i = start-1
#             ed = g.edge_data(e_layer[i])
#             ed.x = x + 1.5
#             ed.y = sum(g.vertex_data(v).y for v in ed.s)/len(ed.s) if len(ed.s) != 0 else 0
#             pad = ed.box_size() * 0.5
#             min_y = pad
#             max_y = -pad

#         for i1 in range(start, end):
#             i0 = end - i1 - 1

#             # place one edge above max_y
#             ed = g.edge_data(e_layer[i0])
#             pad = ed.box_size() * 0.5
#             ed.x = x + 1.5
#             ed.y = min(max_y - pad,
#                        sum(g.vertex_data(v).y for v in ed.s)/len(ed.s)
#                            if len(ed.s) != 0 else 0)
#             max_y = ed.y - pad

#             # place one edge below min_y
#             ed = g.edge_data(e_layer[i1])
#             pad = ed.box_size() * 0.5
#             ed.x = x + 1.5
#             ed.y = max(min_y + pad,
#                        sum(g.vertex_data(v).y for v in ed.s)/len(ed.s)
#                            if len(ed.s) != 0 else 0)
#             min_y = ed.y + pad

#         x += 3.0

#     for i, v in enumerate(v_layers[-1]):
#         vd = g.vertex_data(v)
#         vd.x = x
#         vd.y = i - ((len(v_layers[-1]) - 1) * 0.5)




# def ready_edges(g: Graph, edges: Set[int], v_done: Set[int]) -> Set[int]:
#     ready = set()
#     for e in edges:
#         src = g.source(e)
#         if all(v in v_done for v in src):
#             ready.add(e)
#     return ready
#
# def layer_layout(g: Graph):
#     """A simple hypergraph layout into evenly-spaced layers
#
#     This layout adds identity boxes if wires need to cross more than one layer, and generally places vertices
#     and boxes as early and near the x-axis as possible. It will throw an exception if the given graph is not
#     directed acyclic.
#
#     Currently it assumes graphs are monogamous. Expect strange behaviours if this is not the case.
#     """
#     layer = list(g.inputs())
#     v_done = set(layer)
#     edges = set(g.edges())
#     new_ids = set()
#
#     x = 0
#     while len(edges) > 0:
#         for i, v in enumerate(layer):
#             g.vertex_data(v).x = x
#             g.vertex_data(v).y = i - (len(layer) - 1)/2
#
#         new_layer = []
#         for i, e in enumerate(ready_edges(g, edges, v_done)):
#             src = g.source(e)
#             c = sum(g.vertex_data(v).y for v in src) / len(src) if len(src) != 0 else 0
#             new_layer.append((c, e, g.target(e)))
#             edges.remove(e)
#         if all(e in new_ids for _, e, _ in new_layer):
#             raise ValueError("Could not make progress. Is graph acyclic?")
#
#         # sort according to center coordinate
#         list.sort(new_layer)
#
#         # for c, e, _ in new_layer:
#         #     print("before " + str(e) + ":" + str(g.edge_data(e)) + " center " + str(c))
#
#         min = 0
#         max = 0
#         ctr = -1
#
#         # forward pass, making enough space by shifting things below the x-axis down
#         for i in range(len(new_layer)):
#             c, e, vs = new_layer[i]
#             pad = (g.edge_data(e).box_size() / 2)
#             if ctr == -1:
#                 # the center should be the first place we cross the x-axis or the last element in the list
#                 if c >= 0 or i == len(new_layer) - 1:
#                     ctr = i
#                     max = c - pad
#                     min = c + pad
#             else:
#                 if c < min + pad:
#                     c = min + pad
#                     min = c + pad
#                     new_layer[i] = (c, e, vs)
#
#         # backward pass, shifting things above the x-axis up
#         for i in range(ctr-1, -1, -1):
#             c, e, vs = new_layer[i]
#             pad = (g.edge_data(e).box_size() / 2)
#             if c > max - pad:
#                 c = max - pad
#                 max = c - pad
#                 new_layer[i] = (c, e, vs)
#
#         layer = []
#         for c, e, vs in new_layer:
#             g.edge_data(e).x = x + 1.5
#             g.edge_data(e).y = c
#             # print("after " + str(e) + ":" + str(g.edge_data(e)) + " center " + str(c))
#             for v in vs:
#                 layer.append(v)
#                 v_done.add(v)
#
#         ready = ready_edges(g, edges, v_done)
#         outputs = set(g.outputs())
#         for v in layer:
#             if (v in outputs and len(edges) > 0) or any(e not in ready for e in g.out_edges(v)):
#                 id = g.insert_id_after(v)
#                 edges.add(id)
#                 new_ids.add(id)
#
#         x += 3
#
#     layer = g.outputs()
#     for i, v in enumerate(layer):
#         g.vertex_data(v).x = x
#         g.vertex_data(v).y = i - (len(layer) - 1)/2
#
#     shift = math.floor(x / 2)
#     for v in g.vertices(): g.vertex_data(v).x -= shift
#     for e in g.edges(): g.edge_data(e).x -= shift
#



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sys import prefix
from typing import Iterable, Optional, Tuple
import re
from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent, QTextCursor
from PySide6.QtWidgets import QCompleter, QPlainTextEdit

from .completion import CodeCompletionModel
from .document import ChypDocument
from .highlighter import BG, FG, NO_STATUS


class CodeView(QPlainTextEdit):
    def __init__(self) -> None:
        super().__init__()
        self.setStyleSheet("QPlainTextEdit {background-color: %s; color: %s}" % (BG, FG))
        self.completer = QCompleter(self)
        self.completer.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self.completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
        self.completer.setModelSorting(QCompleter.ModelSorting.CaseSensitivelySortedModel)
        self.completer.setWidget(self)
        self.completion_model = CodeCompletionModel(self.completer)
        # self.completion_model.set_completions(["foo", "bar", "baz"])
        self.completer.setModel(self.completion_model)
        self.completer.activated.connect(self.insert_completion)

    def popup_visible(self) -> bool:
        return self.completer.popup().isVisible()

    def set_completions(self, completions: Iterable[str]) -> None:
        self.completion_model.set_completions(completions)

    def ident_at_cursor(self) -> str:
        cursor = self.textCursor()
        block_pos = cursor.positionInBlock() + 1
        cursor.select(QTextCursor.SelectionType.BlockUnderCursor)
        block = cursor.selectedText()
        # print('full block: "{}"'.format(block))
        block = block[:block_pos]
        # print('chopped block: "{}"'.format(block))

        m = re.search('([a-zA-Z_][\\.a-zA-Z0-9_]*)$', block)
        if m:
            return m.group(1)
        else:
            return ''

    def insert_completion(self, completion: str) -> None:
        if self.completer.widget() is not self:
            return

        # print("inserting completion: " + completion)
        cursor = self.textCursor()
        extra = len(completion) - len(self.completer.completionPrefix())

        if extra != 0:
            cursor.movePosition(QTextCursor.MoveOperation.Left)
            cursor.movePosition(QTextCursor.MoveOperation.EndOfWord)
            cursor.insertText(completion[-extra:])
            self.setTextCursor(cursor)

    def keyPressEvent(self, e: QKeyEvent) -> None:
        if (self.completer.popup().isVisible() and
            e.key() in (Qt.Key.Key_Escape, Qt.Key.Key_Tab, Qt.Key.Key_Backtab)):
            e.ignore()
            return

        complete = e.key() == Qt.Key.Key_Tab or (Qt.KeyboardModifier.ControlModifier in e.modifiers() and e.key() == Qt.Key.Key_Space)
        if self.completer.popup().isVisible() or complete:
            if not complete:
                super().keyPressEvent(e)

            prefix = self.ident_at_cursor()
            cursor = self.textCursor()
            cursor.setPosition(cursor.position() - len(prefix))
            cr = self.cursorRect(cursor)
            # print("prefix: " + prefix)
            if prefix != self.completer.completionPrefix():
                self.completer.setCompletionPrefix(prefix)
                self.completer.popup().setCurrentIndex(self.completer.completionModel().index(0,0))
            cr.setWidth(self.completer.popup().sizeHintForColumn(0) + self.completer.popup().verticalScrollBar().sizeHint().width())
            self.completer.complete(cr)
        else:
            super().keyPressEvent(e)

    def set_current_region(self, region: Optional[Tuple[int,int]], status: int = NO_STATUS) -> None:
        doc = self.document()

        if isinstance(doc, ChypDocument):
            self.blockSignals(True)
            doc.highlighter.set_current_region(region, status)
            self.blockSignals(False)

    def add_line_below(self, text: str) -> None:
        doc = self.document()

        if isinstance(doc, ChypDocument):
            cursor = self.textCursor()
            if doc.highlighter.region:
                cursor.setPosition(doc.highlighter.region[1])
            cursor.movePosition(QTextCursor.MoveOperation.EndOfLine)
            cursor.insertText('\n' + text)
            self.setTextCursor(cursor)





#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Callable, Dict, Optional, Tuple
from PySide6.QtCore import QByteArray, QFileInfo, QObject, QThread, QTimer, Qt, QSettings
from PySide6.QtGui import QTextCursor
from PySide6.QtWidgets import QHBoxLayout, QSplitter, QTreeView, QVBoxLayout, QWidget

from ..layout import convex_layout
from ..graph import Graph
from ..state import RewriteState, State
from ..term import graph_to_term
from ..matcher import match_rule
from ..rewrite import rewrite

from .errorlist import ErrorListModel
from .graphview import GraphView
from .codeview import CodeView
from .document import ChypDocument
from .highlighter import STATUS_GOOD, STATUS_BAD

class Editor(QWidget):
    def __init__(self) -> None:
        super().__init__()
        conf = QSettings('chyp', 'chyp')

        self.setLayout(QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0,0,0,0)

        # save splitter position
        self.splitter = QSplitter(Qt.Orientation.Vertical)
        self.layout().addWidget(self.splitter)

        self.state = State()

        self.lhs_view = GraphView()
        self.rhs_view = GraphView()
        self.rhs_view.setVisible(False)

        self.graph_panel = QWidget(self)
        self.graph_panel.setLayout(QHBoxLayout())
        self.graph_panel.layout().addWidget(self.lhs_view)
        self.graph_panel.layout().addWidget(self.rhs_view)
        self.splitter.addWidget(self.graph_panel)

        self.code_view = CodeView()
        self.doc = ChypDocument(self)
        self.code_view.setDocument(self.doc)
        # self.doc.documentReplaced.connect(self.reset_state)

        self.splitter.addWidget(self.code_view)

        self.error_view = QTreeView()
        self.error_view.setIndentation(0)
        self.error_view.setModel(ErrorListModel())
        self.splitter.addWidget(self.error_view)

        splitter_state = conf.value("editor_splitter_state")
        if splitter_state and isinstance(splitter_state, QByteArray): self.splitter.restoreState(splitter_state)

        self.code_view.cursorPositionChanged.connect(self.show_at_cursor)
        self.code_view.textChanged.connect(self.invalidate_text)
        self.parsed = True

        # keep a cache of graphs that have already been laid out
        self.graph_cache : Dict[int, Tuple[Graph, Optional[Graph]]] = dict()

        # keep a revision count, so we don't trigger parsing until the user stops typing for a bit
        self.revision = 0

        # index of the part of the parsed document we're currently looking at
        self.current_part = -1


    def title(self) -> str:
        if self.doc.file_name:
            fi = QFileInfo(self.doc.file_name)
            title = fi.fileName()
        else:
            title = 'Untitled'

        if self.doc.isModified():
            title += '*'

        return title

    def reset_state(self) -> None:
        cursor = self.code_view.textCursor()
        cursor.setPosition(0)
        self.code_view.setTextCursor(cursor)

    def invalidate_text(self) -> None:
        self.parsed = False
        self.current_part = -1
        self.graph_cache = dict()
        self.code_view.set_current_region(None)
        self.revision += 1

        def update(r: int) -> Callable:
            def f() -> None:
                if r == self.revision:
                    self.update_state()
            return f

        QTimer.singleShot(200, update(self.revision))
        # self.update_state(sync=False)

    def next_part(self, step:int=1) -> None:
        if not self.parsed: return

        cursor = self.code_view.textCursor()
        pos = cursor.position()
        p = self.state.part_with_index_at(pos)
        i = p[0] if p else 0
        i += step

        if i >= 0 and i < len(self.state.parts):
            p1 = self.state.parts[i]
            cursor.setPosition(p1[1])
            self.code_view.setTextCursor(cursor)
            
    def show_errors(self) -> None:
        conf = QSettings('chyp', 'chyp')
        error_panel_size = conf.value('error_panel_size', 100)
        if isinstance(error_panel_size, str):
            error_panel_size = int(error_panel_size)
        if not isinstance(error_panel_size, int) or error_panel_size == 0:
            error_panel_size = 100
        
        sizes = self.splitter.sizes()
        if sizes[2] == 0:
            sizes[2] = error_panel_size
            if sizes[1] >= error_panel_size + 50:
                sizes[1] -= error_panel_size
            else:
                sizes[0] -= error_panel_size
        else:
            conf.setValue('error_panel_size', sizes[2])
            sizes[1] += sizes[2]
            sizes[2] = 0
        self.splitter.setSizes(sizes)

    def show_at_cursor(self) -> None:
        if not self.parsed: return
        pos = self.code_view.textCursor().position()
        p = self.state.part_with_index_at(pos)
        if not p: return

        i, part = p
        # print(part)

        if i == self.current_part: return
        else: self.current_part = i

        self.code_view.set_current_region((part[0], part[1]))
        if part[2] in ('let','gen') and part[3] in self.state.graphs:
            if i not in self.graph_cache:
                g = self.state.graphs[part[3]].copy()
                convex_layout(g)
                self.graph_cache[i] = (g, None)
            else:
                g, _ = self.graph_cache[i]
            self.rhs_view.setVisible(False)
            self.lhs_view.set_graph(g)
        elif part[2] == 'rule' and part[3] in self.state.rules:
            if i not in self.graph_cache:
                lhs = self.state.rules[part[3]].lhs.copy()
                rhs = self.state.rules[part[3]].rhs.copy()
                convex_layout(lhs)
                convex_layout(rhs)
                self.graph_cache[i] = (lhs, rhs)
            else:
                lhs, rhs0 = self.graph_cache[i]
                if not rhs0: raise ValueError("Rule in graph_cache should have RHS")
                rhs = rhs0
            self.rhs_view.setVisible(True)
            self.lhs_view.set_graph(lhs)
            self.rhs_view.set_graph(rhs)
        elif part[2] == 'rewrite' and part[3] in self.state.rewrites:
            rw = self.state.rewrites[part[3]]
            if not rw.stub:
                if rw.status == RewriteState.UNCHECKED:
                    rw.status = RewriteState.CHECKING
                    def check_finished(i: int) -> Callable:
                        def f() -> None:
                            self.graph_cache.pop(i, None)
                            self.show_at_cursor()
                        return f

                    check_thread = CheckThread(rw, self)
                    check_thread.finished.connect(check_finished(i))
                    check_thread.start()

            if i not in self.graph_cache:
                lhs = rw.lhs.copy() if rw.lhs else Graph()
                rhs = rw.rhs.copy() if rw.rhs else Graph()
                convex_layout(lhs)
                convex_layout(rhs)
                self.graph_cache[i] = (lhs, rhs)
            else:
                lhs, rhs0 = self.graph_cache[i]
                if not rhs0: raise ValueError("Rewrite step in graph_cache should have RHS")
                rhs = rhs0

            if rw.status == RewriteState.VALID:
                self.code_view.set_current_region((part[0], part[1]), status=STATUS_GOOD)
            elif rw.status == RewriteState.INVALID:
                self.code_view.set_current_region((part[0], part[1]), status=STATUS_BAD)

            self.rhs_view.setVisible(True)
            self.lhs_view.set_graph(lhs)
            if rhs: self.rhs_view.set_graph(rhs)
        else:
            self.rhs_view.setVisible(False)
            self.lhs_view.set_graph(Graph())

    def next_rewrite_at_cursor(self) -> None:
        self.update_state()
        if not self.parsed: return

        pos = self.code_view.textCursor().position()
        part = self.state.part_at(pos)
        if part and part[2] == 'rewrite' and part[3] in self.state.rewrites:
            rw = self.state.rewrites[part[3]]
            if rw.rule and rw.lhs:
                start, end = rw.term_pos
                text = self.code_view.toPlainText()
                term = text[start:end]
                seen = set([term])

                found_prev = (term == '?')
                rw_term = None
                for m in match_rule(rw.rule, rw.lhs):
                    t = graph_to_term(rewrite(rw.rule, m))
                    if found_prev and not t in seen:
                        rw_term = t
                        break
                    elif not rw_term:
                        rw_term = t

                    seen.add(t)
                    found_prev = (term == t)

                if rw_term:
                    cursor = self.code_view.textCursor()
                    cursor.clearSelection()
                    cursor.setPosition(start)
                    cursor.setPosition(end, mode=QTextCursor.MoveMode.KeepAnchor)

                    # self.blockSignals(True)
                    cursor.insertText(rw_term)
                    self.code_view.setTextCursor(cursor)
                    # self.blockSignals(False)
                    # self.update_state()

    def repeat_step_at_cursor(self) -> None:
        self.update_state()
        pos = self.code_view.textCursor().position()
        part = self.state.part_at(pos)
        if part and part[2] == 'rewrite' and part[3] in self.state.rewrites:
            rule = self.state.rewrites[part[3]].rule

            if rule:
                self.code_view.add_line_below('  = ? by ' + rule.name)
                self.update_state()
                self.next_rewrite_at_cursor()

    def update_state(self) -> None:
        self.state = State()
        self.code_view.set_current_region(None)
        
        self.state.update(self.doc.toPlainText(), self.doc.file_name)
        model = self.error_view.model()
        if isinstance(model, ErrorListModel):
            model.set_errors(self.state.errors)

        if len(self.state.errors) == 0:
            self.parsed = True
            self.code_view.set_completions(self.state.rules.keys())
            self.show_at_cursor()

    def import_at_cursor(self) -> str:
        p = self.state.part_at(self.code_view.textCursor().position())
        if p and p[2] == 'import':
            return p[3]
        else:
            return ''

class CheckThread(QThread):
    def __init__(self, rw: RewriteState, parent: Optional[QObject] = None) -> None:
        super().__init__(parent)
        self.rw = rw

    def run(self) -> None:
        self.rw.check()

# class UpdateStateThread(QThread):
#     def __init__(self, state: State, code: str, parent: Optional[QObject] = None) -> None:
#         super().__init__(parent)
#         self.state = state
#         self.code = code

#     def run(self) -> None:
#         self.msleep(300)
#         self.state.update(self.code)



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict
from PySide6.QtCore import QSettings
from PySide6.QtGui import QColor, QPalette
from PySide6.QtWidgets import QApplication

PALETTES: Dict[str, Dict[str,str]] = dict()

PALETTES['catppuccin_macchiato'] = {
    'rosewater': '#f4dbd6',
    'flamingo':  '#f0c6c6',
    'pink':      '#f5bde6',
    'mauve':     '#c6a0f6',
    'red':       '#ed8796',
    'maroon':    '#ee99a0',
    'peach':     '#f5a97f',
    'yellow':    '#eed49f',
    'green':     '#a6da95',
    'teal':      '#8bd5ca',
    'sky':       '#91d7e3',
    'sapphire':  '#7dc4e4',
    'blue':      '#8aadf4',
    'lavender':  '#b7bdf8',
    'text':      '#cad3f5',
    'subtext1':  '#b8c0e0',
    'subtext0':  '#a5adcb',
    'overlay2':  '#939ab7',
    'overlay1':  '#8087a2',
    'overlay0':  '#6e738d',
    'surface2':  '#5b6078',
    'surface1':  '#494d64',
    'surface0':  '#363a4f',
    'base':      '#24273a',
    'mantle':    '#1e2030',
    'crust':     '#181926',

    'green_bg':  '#36504f', # non-standard
    'red_bg':    '#863a4f', # non-standard
}

PALETTES['catppuccin_latte'] = {
    'rosewater': '#dc8a78',
    'flamingo':  '#dd7878',
    'pink':      '#ea76cb',
    'mauve':     '#8839ef',
    'red':       '#d20f39',
    'maroon':    '#e64553',
    'peach':     '#fe640b',
    'yellow':    '#df8e1d',
    'green':     '#40a02b',
    'teal':      '#179299',
    'sky':       '#04a5e5',
    'sapphire':  '#209fb5',
    'blue':      '#1e66f5',
    'lavender':  '#7287fd',
    'text':      '#4c4f69',
    'subtext1':  '#5c5f77',
    'subtext0':  '#6c6f85',
    'overlay2':  '#7c7f93',
    'overlay1':  '#8c8fa1',
    'overlay0':  '#9ca0b0',
    'surface2':  '#acb0be',
    'surface1':  '#bcc0cc',
    'surface0':  '#ccd0da',
    'base':      '#eff1f5',
    'mantle':    '#e6e9ef',
    'crust':     '#dce0e8',

    'green_bg':  '#bbffcc', # non-standard
    'red_bg':    '#ff9999', # non-standard
}

THEMES: Dict[str, Dict[str,str]] = dict()
for t in ('catppuccin_macchiato', 'catppuccin_latte'):
    THEMES[t] = {
      'bg': PALETTES[t]['base'],
      'fg': PALETTES[t]['text'],
      'fg_bright': PALETTES[t]['lavender'],
      'bg_good': PALETTES[t]['green_bg'],
      'bg_bad': PALETTES[t]['red_bg'],
      'bg_alt': PALETTES[t]['crust'],
      'bg_button': PALETTES[t]['mantle'],
      'fg_button': PALETTES[t]['text'],
      'fg_link': PALETTES[t]['blue'],
      'bg_highlight': PALETTES[t]['blue'],
      'fg_highlight': PALETTES[t]['crust'],

      'bg_sel': PALETTES[t]['surface0'], 
      'fg_keyword': PALETTES[t]['blue'],
      'fg_ident': PALETTES[t]['sky'], 
      'fg_num': PALETTES[t]['red'],
      'fg_op': PALETTES[t]['yellow'],
      'fg_comment': PALETTES[t]['overlay1'],
      'fg_string': PALETTES[t]['green']
    }

def current_theme() -> Dict[str,str]:
    conf = QSettings('chyp', 'chyp')

    theme_name = conf.value('theme')
    if not theme_name or not isinstance(theme_name, str):
        theme_name = 'catppuccin_macchiato'

    return THEMES[theme_name]

def apply_theme() -> None:
    theme = current_theme()

    QApplication.setStyle("Fusion")
    # Now use a palette to switch to theme colors:
    palette = QPalette()
    palette.setColor(QPalette.ColorRole.Window, QColor(theme['bg']))
    palette.setColor(QPalette.ColorRole.WindowText, QColor(theme['fg']))
    palette.setColor(QPalette.ColorRole.Base, QColor(theme['bg']))
    palette.setColor(QPalette.ColorRole.AlternateBase, QColor(theme['bg_alt']))
    palette.setColor(QPalette.ColorRole.ToolTipBase, QColor(theme['fg']))
    palette.setColor(QPalette.ColorRole.ToolTipText, QColor(theme['fg']))
    palette.setColor(QPalette.ColorRole.Text, QColor(theme['fg']))
    palette.setColor(QPalette.ColorRole.Button, QColor(theme['bg_button']))
    palette.setColor(QPalette.ColorRole.ButtonText, QColor(theme['fg_button']))
    palette.setColor(QPalette.ColorRole.BrightText, QColor(theme['fg_bright']))
    palette.setColor(QPalette.ColorRole.Link, QColor(theme['fg_link']))
    palette.setColor(QPalette.ColorRole.Highlight, QColor(theme['bg_highlight']))
    palette.setColor(QPalette.ColorRole.HighlightedText, QColor(theme['fg_highlight']))
    QApplication.setPalette(palette)



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import itertools
from typing import Optional, List, Tuple

from PySide6.QtCore import Qt, QPointF
from PySide6.QtGui import QBrush, QColor, QFont, QPainter, QPainterPath, QPen, QTransform, QUndoStack
from PySide6.QtWidgets import (QGraphicsEllipseItem, QGraphicsItem,
                               QGraphicsPathItem, QGraphicsRectItem, QGraphicsScene,
                               QGraphicsSceneMouseEvent, QStyleOptionGraphicsItem, QWidget)

from ..graph import Graph

SCALE = 50.0

BOX_WIDTH = 1.1

class EItem(QGraphicsRectItem):
    def __init__(self, g: Graph, e: int) -> None:
        super().__init__(-0.5 * BOX_WIDTH * SCALE, -0.8 * SCALE, BOX_WIDTH * SCALE, 1.6 * SCALE)
        self.g = g
        self.e = e
        ed = g.edge_data(e)
        self.bg = ed.bg
        self.fg = ed.fg
        self.value = ed.value
        self.num_s = len(ed.s)
        self.num_t = len(ed.t)
        self.is_id = isinstance(ed.value, str) and str(ed.value) == 'id'

        self.setPos(ed.x * SCALE, ed.y * SCALE)

        if self.is_id:
            self.setRect(-0.2 * SCALE, -0.2 * SCALE, 0.4 * SCALE, 0.4 * SCALE)
            self.setPen(QPen(QColor(200,200,200)))
            self.setBrush(QBrush(QColor(200,200,200)))
            self.setVisible(False)
        else:
            if self.num_s <= 1 and self.num_t <= 1:
                self.setRect(-0.5 * BOX_WIDTH * SCALE, -0.4 * SCALE, BOX_WIDTH * SCALE, 0.8 * SCALE)

            if self.bg != '':
                self.setBrush(QBrush(QColor(self.bg)))
            else:
                self.setBrush(QBrush(QColor(200,200,255)))

            pen = QPen(QColor(0,0,0))
            if ed.highlight:
                pen.setWidth(3)
            self.setPen(pen)

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget: Optional[QWidget]=None) -> None:
        super().paint(painter, option, widget)

        if self.fg != '':
            painter.setPen(QPen(QColor(self.fg)))
        painter.setFont(QFont("sans", 11))
        painter.drawText(self.boundingRect(), Qt.AlignmentFlag.AlignCenter, str(self.value)) # type:ignore

class VItem(QGraphicsEllipseItem):
    def __init__(self, g: Graph, v: int, eitem: Optional[EItem]=None, i: int=-1) -> None:
        super().__init__(-0.0625 * SCALE, -0.0625 * SCALE, 0.125 * SCALE, 0.125 * SCALE)
        self.setVisible(False)
        self.g = g
        self.v = v
        self.eitem = eitem
        self.i = i
        vd = g.vertex_data(v)
        self.setPos(vd.x * SCALE, vd.y * SCALE)
        self.setBrush(QBrush(QColor(0,0,0)))

    def refresh(self):
        if self.eitem and self.i != -1:
            x_shift = 0.7 * SCALE
            if self.eitem.num_t == 1:
                y_shift = 0
            else:
                y_shift = ((self.i / (self.eitem.num_t - 1)) - 0.5) * SCALE
            p = self.eitem.pos()
            self.setPos(p.x() + x_shift, p.y() + y_shift)

class TItem(QGraphicsPathItem):
    def __init__(self, vitem: VItem, eitem: EItem, i: int, src: bool) -> None:
        super().__init__()
        g = vitem.g
        self.vitem = vitem
        self.eitem = eitem
        if (g.vertex_data(vitem.v).highlight or
            g.edge_data(eitem.e).highlight):
            pen = QPen()
            pen.setWidth(3)
            self.setPen(pen)
        self.i = i
        self.src = src
        self.refresh()

    def refresh(self) -> None:
        path = QPainterPath()

        x_shift = (0 if self.eitem.is_id else 0.5 * BOX_WIDTH) * SCALE

        if self.src:
            if self.eitem.num_s == 1:
                y_shift = 0.0
            else:
                y_shift = ((self.i / (self.eitem.num_s - 1)) - 0.5) * SCALE

            p1x = self.vitem.pos().x()
            p1y = self.vitem.pos().y()
            p2x = self.eitem.pos().x() - x_shift
            p2y = self.eitem.pos().y() + y_shift
        else:
            if self.eitem.num_t == 1:
                y_shift = 0
            else:
                y_shift = ((self.i / (self.eitem.num_t - 1)) - 0.5) * SCALE

            p1x = self.eitem.pos().x() + x_shift
            p1y = self.eitem.pos().y() + y_shift
            p2x = self.vitem.pos().x()
            p2y = self.vitem.pos().y()

        dx = abs(p1x - p2x)
        path.moveTo(p1x, p1y)
        path.cubicTo(p1x + dx * 0.4, p1y,
                     p2x - dx * 0.4, p2y,
                     p2x, p2y)
        self.setPath(path)
        self.update(-2000,-2000,4000,4000)

class GraphScene(QGraphicsScene):
    def __init__(self) -> None:
        super().__init__()
        self.undo_stack = QUndoStack(self)

        self.setSceneRect(-2000, -2000, 4000, 4000)
        self.setBackgroundBrush(QBrush(QColor(255,255,255)))
        self.drag_start = QPointF(0,0)
        self.drag_items: List[Tuple[QGraphicsItem, QPointF]] = []

    def set_graph(self, g: Graph) -> None:
        self.g = g
        self.clear()
        self.add_items()
        self.invalidate()

    def add_items(self) -> None:
        vi = {}
        ei = {}
        for e in self.g.edges():
            ei[e] = EItem(self.g, e)
            self.addItem(ei[e])

        for v in itertools.chain(self.g.inputs(), self.g.outputs()):
            vi[v] = VItem(self.g, v)
            self.addItem(vi[v])

        for e in self.g.edges():
            ed = self.g.edge_data(e)
            for i, v in enumerate(ed.t):
                if not v in vi:
                    vi[v] = VItem(self.g, v, ei[e], i)
                    self.addItem(vi[v])

        for e in self.g.edges():
            ed = self.g.edge_data(e)
            for i, v in enumerate(ed.s):
                ti = TItem(vi[v], ei[e], i, src=True)
                self.addItem(ti)
            for i, v in enumerate(ed.t):
                ti = TItem(vi[v], ei[e], i, src=False)
                self.addItem(ti)

    def mousePressEvent(self, e: QGraphicsSceneMouseEvent) -> None:
        super().mousePressEvent(e)
        
        self.drag_start = e.scenePos()

        for it in self.items(e.scenePos(), deviceTransform=QTransform()):
            if it and (isinstance(it, EItem) or isinstance(it, VItem)):
                self.drag_items = [(it, it.scenePos())]
                break

    def mouseMoveEvent(self, e: QGraphicsSceneMouseEvent) -> None:
        p = e.scenePos()
        grid_size = SCALE / 8
        dx = round((p.x() - self.drag_start.x())/grid_size) * grid_size
        dy = round((p.y() - self.drag_start.y())/grid_size) * grid_size

        # move the items that have been dragged
        for it,pos in self.drag_items:
            it.setPos(QPointF(pos.x() + dx, pos.y() + dy))

        for it in self.items():
            # update positions for any vertices and tentacles attached to dragged items
            if isinstance(it, TItem):
                for it1,_ in self.drag_items:
                    if it.vitem == it1 or it.eitem == it1 or it.vitem.eitem == it1:
                        it.vitem.refresh()
                        it.refresh()
                        break

            # if the mouse is near a vertex or identity box, make it visible
            if isinstance(it, VItem) or (isinstance(it, EItem) and it.is_id):
                if abs(it.pos().x() - p.x()) < 10 and abs(it.pos().y() - p.y()) < 10:
                    it.setVisible(True)
                else:
                    it.setVisible(False)


    def mouseReleaseEvent(self, _: QGraphicsSceneMouseEvent) -> None:
        self.drag_items = []




# type: ignore
import setuptools

# read description from README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


data_files = []

setuptools.setup(
    name="chyp",
    version="0.4.1",
    author="Aleks Kissinger",
    author_email="aleks0@gmail.com",
    description="An interactive theorem prover for string diagrams",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akissinger/chyp",
    project_urls={
        "Bug Tracker": "https://github.com/akissinger/chyp/issues",
    },
    license="Apache2",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=["chyp", "chyp.gui"],
    package_data={'': ['*.svg']},
    data_files=data_files,
    install_requires=["PySide6>=6.4.3", "lark>=1.1.5", "cvxpy>=1.3.1"],
    python_requires=">=3.7",
    entry_points={'console_scripts': 'chyp=chyp.gui.app:main'},
)



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Iterable, List, Union
from PySide6.QtCore import QAbstractListModel, QObject, QPersistentModelIndex, Qt, QAbstractItemModel, QModelIndex
from PySide6.QtGui import QFont

class CodeCompletionModel(QAbstractListModel):
    """A model containing a list of completions, with line numbers"""
    completions: List[str]

    def __init__(self, parent: QObject) -> None:
        super().__init__(parent)
        self.completions = []

    def set_completions(self, completions: Iterable[str]) -> None:
        self.beginResetModel()
        self.completions = sorted(completions)
        self.endResetModel()

    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: int=Qt.ItemDataRole.DisplayRole) -> Any:
        """Overrides `QAbstractItemModel.data` to populate a view with completions"""

        if index.row() >= len(self.completions) or index.column() >= 3:
            return None

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if index.column() == 0:
                return self.completions[index.row()]
        elif role == Qt.ItemDataRole.FontRole:
            return QFont("monospace", 14)

    def rowCount(self, index: Union[QModelIndex, QPersistentModelIndex]=QModelIndex()) -> int:
        """The number of rows"""
        if not index or not index.isValid(): return len(self.completions)
        else: return 0





#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os.path
from typing import Any, Dict, List, Optional, Tuple
from lark import Lark, ParseTree, Transformer, UnexpectedInput, v_args
from lark.tree import Meta

from .graph import Graph, GraphError, gen, perm, identity
from .rule import Rule, RuleError

GRAMMAR = Lark("""
    start : statement*
    ?statement : import_statement | gen | let | def_statement | rule | rewrite | show
    gen : "gen" var ":" num "->" num [ gen_color ]
    def_statement : "def" var "=" term [ gen_color ]
    gen_color : "\\\"" color "\\\"" | "\\\"" color "\\\"" "\\\"" color "\\\""
    let : "let" var "=" term
    rule : "rule" var ":" term (eq | le) term
    rewrite : "rewrite" [converse] var ":" term rewrite_part*
    rewrite_part : (eq | le) term_hole [ "by" [ converse ] rule_ref ]
    converse : "-"
    ?term  : par_term | seq
    ?par_term : "(" term ")" | par | perm | id | id0 | term_ref
    par : par_term "*" par_term
    seq : term ";" term
    perm : "sw" [ "[" num ("," num)* "]" ]
    id : "id"
    id0 : "id0"
    show : "show" rule_ref

    import_statement : "import" module_name [ "as" var ] [ "(" import_let ("," import_let)* ")" ]
    import_let : var "=" term

    eq : "=" | "=="
    le : "<=" | "~>"
    num : INT
    module_name : IDENT
    var : IDENT
    term_ref : IDENT
    rule_ref : IDENT
    term_hole : term | "?"
    color : HEXDIGIT HEXDIGIT HEXDIGIT HEXDIGIT HEXDIGIT HEXDIGIT
    IDENT: ("_"|LETTER) ("_"|"."|LETTER|DIGIT)*

    %import common.LETTER
    %import common.DIGIT
    %import common.HEXDIGIT
    %import common.INT
    %import common.WS
    %import common.SH_COMMENT
    %ignore WS
    %ignore SH_COMMENT
    """,
    parser='lalr',
    propagate_positions=True,
    maybe_placeholders=True)


source_cache: Dict[str, Tuple[float,str]] = dict()

class ChypParseData(Transformer):
    def __init__(self, namespace: str, file_name: str) -> None:
        self.namespace = namespace
        self.file_name = file_name
        self.import_depth = 0

        self.graphs: Dict[str, Graph] = dict()
        self.rules: Dict[str, Rule] = {'refl': Rule(Graph(), Graph(), name="refl")}
        self.rewrites: Dict[str, Tuple[int, int, bool, Optional[Rule],
                                       Optional[Graph], Optional[Graph],
                                       Optional[Graph], Optional[Graph]]] = dict()
        self.errors: List[Tuple[str, int, str]] = list()
        self.parts: List[Tuple[int, int, str, str]] = list()
        self.parsed = False
    
    # def start(self, items: List[List[Tuple[int,int,str,str]]]):
    #     pass
        
    def var(self, items: List[Any]) -> str:
        s = str(items[0])
        return self.namespace + '.' + s if self.namespace else s

    def module_name(self, items: List[Any]) -> str:
        return str(items[0])
    
    def num(self, items: List[Any]) -> int:
        return int(items[0])

    def id(self, _: List[Any]) -> Graph:
        return identity()

    def id0(self, _: List[Any]) -> Graph:
        return Graph()

    def eq(self, _: List[Any]) -> bool:
        return True

    def le(self, _: List[Any]) -> bool:
        return False

    @v_args(meta=True)
    def perm(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        try:
            if items[0] is None:
                return perm([1,0])
            else:
                return perm([int(i) for i in items])
        except GraphError as e:
            self.errors.append((self.file_name, meta.line, str(e)))
        return None
    
    @v_args(meta=True)
    def term_ref(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        s = str(items[0])
        if self.namespace:
            s = self.namespace + '.' + s

        if s in self.graphs:
            return self.graphs[s]
        else:
            self.errors.append((self.file_name, meta.line, 'Undefined term: ' + s))
            return None

    @v_args(meta=True)
    def rule_ref(self, meta: Meta, items: List[Any]) -> Optional[Rule]:
        s = str(items[0])
        if self.namespace and s != 'refl':
            s = self.namespace + '.' + s

        if s in self.rules:
            return self.rules[s]
        else:
            self.errors.append((self.file_name, meta.line, 'Undefined rule: ' + s))
            return None
    
    def par(self, items: List[Any]) -> Optional[Graph]:
        if items[0] and items[1]:
            return items[0] * items[1]
        else:
            return None
    
    @v_args(meta=True)
    def seq(self, meta: Meta, items: List[Any]) -> Optional[Graph]:
        if items[0] and items[1]:
            g = None
            try:
                g = items[0] >> items[1]
            except GraphError as e:
                self.errors.append((self.file_name, meta.line, str(e)))
            return g
        else:
            return None

    @v_args(meta=True)
    def show(self, meta: Meta, items: List[Any]) -> None:
        rule = items[0]
        if rule:
            self.parts.append((meta.start_pos, meta.end_pos, 'rule', rule.name))

    @v_args(meta=True)
    def gen(self, meta: Meta, items: List[Any]) -> None:
        name = items[0]
        arity = items[1]
        coarity = items[2]
        (fg,bg) = items[3] if items[3] else ('','')
        if not name in self.graphs:
            self.graphs[name] = gen(name, arity, coarity, fg, bg)
        else:
            g = self.graphs[name]
            inp = len(g.inputs())
            outp = len(g.outputs())
            if inp != arity or outp != coarity:
                self.errors.append((self.file_name, meta.line, "Term '{}' already defined with incompatible type {} -> {}.".format(name, inp, outp)))
        self.parts.append((meta.start_pos, meta.end_pos, 'gen', name))
        
    @v_args(meta=True)
    def let(self, meta: Meta, items: List[Any]) -> None:
        name, graph = items
        if not name in self.graphs:
            if graph:
                self.graphs[name] = graph
        else:
            self.errors.append((self.file_name, meta.line, "Term '{}' already defined.".format(name)))
        self.parts.append((meta.start_pos, meta.end_pos, 'let', name))

    @v_args(meta=True)
    def rule(self, meta: Meta, items: List[Any]) -> None:
        name, lhs, invertible, rhs = items
        if not name in self.rules:
            if lhs and rhs:
                try:
                    self.rules[name] = Rule(lhs, rhs, name, invertible)
                except RuleError as e:
                    self.errors.append((self.file_name, meta.line, str(e)))
        else:
            self.errors.append((self.file_name, meta.line, "Rule '{}' already defined.".format(name)))
        self.parts.append((meta.start_pos, meta.end_pos, 'rule', name))

    @v_args(meta=True)
    def def_statement(self, meta: Meta, items: List[Any]) -> None:
        name = items[0]
        graph = items[1]
        (fg,bg) = items[2] if items[2] else ('','')

        rule_name = name + '_def'
        if not rule_name in self.rules:
            if graph:
                arity = len(graph.inputs())
                coarity = len(graph.outputs())

                if not name in self.graphs:
                    lhs = gen(name, arity, coarity, fg, bg)
                    self.graphs[name] = lhs
                    self.rules[rule_name] = Rule(lhs, graph, rule_name, True)
                else:
                    lhs = self.graphs[name]
                    inp = len(lhs.inputs())
                    outp = len(lhs.outputs())
                    if inp == arity or outp == coarity:
                        self.rules[rule_name] = Rule(lhs, graph, rule_name, True)
                    else:
                        self.errors.append((self.file_name, meta.line, "Term '{}' already defined with incompatible type {} -> {}.".format(name, inp, outp)))

        else:
            self.errors.append((self.file_name, meta.line, "Rule '{}' already defined.".format(name, rule_name)))
        self.parts.append((meta.start_pos, meta.end_pos, 'rule', rule_name))

    def gen_color(self, items: List[Any]) -> Tuple[str,str]:
        return (items[1], items[0]) if len(items) == 2 else ('', items[0])

    def color(self, items: List[Any]) -> str:
        return '#' + ''.join([str(it) for it in items])

    @v_args(meta=True)
    def import_statement(self, meta: Meta, items: List[Any]) -> None:
        mod = items[0]
        namespace = items[1] or ''
        import_lets = items[2:]

        for il in import_lets:
            if il:
                name, graph = il
                name = namespace + '.' + name
                if not name in self.graphs:
                    if graph:
                        self.graphs[name] = graph
                else:
                    self.errors.append((self.file_name, meta.line, "Term '{}' already defined.".format(name)))

        file_name = module_filename(mod, self.file_name)
        try:
            parse(file_name=file_name, namespace=namespace, parent=self)
        except FileNotFoundError:
            self.errors.append((self.file_name, meta.line, 'File not found: {}'.format(file_name)))

        self.parts.append((meta.start_pos, meta.end_pos, 'import', file_name))

    def import_let(self, items: List[Any]) -> Tuple[str, Graph]:
        return (items[0], items[1])

    @v_args(meta=True)
    def rewrite(self, meta: Meta, items: List[Any]) -> None:
        converse = True if items[0] else False
        base_name = items[1]
        name = '-' + base_name if converse else base_name

        term = items[2]
        rw_parts = items[3:]

        if len(rw_parts) == 0:
            self.parts.append((meta.start_pos, meta.end_pos, "rewrite", name))
            self.rewrites[name] = (0,0,False,None,term,None,None,None)
        else:
            start = meta.start_pos
            lhs = term
            rhs = None
            all_equiv = True

            if converse and base_name in self.rules:
                lhs_match = self.rules[base_name].rhs
                rhs_match = self.rules[base_name].lhs
            else:
                lhs_match = None
                rhs_match = None

            last_i = len(rw_parts)-1
            for i, rw_part in enumerate(rw_parts):
                end, t_start, t_end, equiv, rule, rhs = rw_part
                all_equiv = all_equiv and equiv
                self.rewrites[name + ":" + str(i)] = (t_start, t_end, equiv, rule, lhs, rhs,
                                                      lhs_match if i == 0 else None,
                                                      rhs_match if i == last_i else None)
                lhs = rhs.copy() if rhs else None
                self.parts.append((start, end, "rewrite", name + ":" + str(i)))
                start = end
            if term and rhs:
                try:
                    if converse:
                        if base_name in self.rules:
                            self.rules[base_name].equiv = True
                        else:
                            self.errors.append((self.file_name, meta.line, "Trying to prove converse for unknown rule: " + base_name))
                    else:
                        if not name in self.rules:
                            self.rules[name] = Rule(term, rhs, name=name, equiv=all_equiv)
                        else:
                            self.errors.append((self.file_name, meta.line, "Rule '{}' already defined.".format(name)))
                except RuleError as e:
                    self.errors.append((self.file_name, meta.line, str(e)))

    @v_args(meta=True)
    def rewrite_part(self, meta: Meta, items: List[Any]) -> Tuple[int, int, int, bool, Optional[Rule], Optional[Graph]]:
        equiv = items[0]
        t_start,t_end,rhs = items[1]
        converse = True if items[2] else False

        rule = items[3] if items[3] else self.rules['refl']
        if rule and converse:
            rule = rule.converse()

        if equiv and rule and not rule.equiv:
            rule = None

        return (meta.end_pos, t_start, t_end, equiv, rule, rhs)


    @v_args(meta=True)
    def term_hole(self, meta: Meta, items: List[Any]) -> Tuple[int, int, Optional[Graph]]:
        t = items[0] if len(items) != 0 else None
        return (meta.start_pos, meta.end_pos, t)


def module_filename(name: str, current_file: str) -> str:
    return os.path.join(os.path.dirname(current_file), *name.split('.')) + '.chyp'

# cache parse trees for imported files and only re-parse if the file changes
parse_cache: Dict[str, Tuple[float, ParseTree]] = dict()

def parse(code: str='', file_name: str='', namespace: str='', parent: Optional[ChypParseData] = None) -> ChypParseData:
    global parse_cache

    if parent and parent.namespace:
        if namespace != '':
            namespace = parent.namespace + '.' + namespace
        else:
            namespace = parent.namespace

    parse_data = ChypParseData(namespace, file_name)

    if parent:
        parse_data.graphs = parent.graphs
        parse_data.rules = parent.rules
        parse_data.errors = parent.errors
        parse_data.import_depth = parent.import_depth + 1
        if parse_data.import_depth > 255:
            parse_data.errors += [(parent.file_name, -1, "Maximum import depth (255) exceeded. Probably a cyclic import.")]

    try:
        if file_name and not code:
            mtime = os.path.getmtime(file_name)
            if file_name in parse_cache and parse_cache[file_name][0] == mtime:
                tree = parse_cache[file_name][1]
            else:
                with open(file_name) as f:
                    tree = GRAMMAR.parse(f.read())
                parse_cache[file_name] = (mtime, tree)
        else:
            tree = GRAMMAR.parse(code)
        parse_data.transform(tree)
        parse_data.parsed = True
    except UnexpectedInput as e:
        msg = 'Parse error: '
        e_lines = e.get_context(code).splitlines()
        if len(e_lines) >= 2:
            parse_data.errors += [(file_name, e.line, msg + e_lines[0] + '\n' + len(msg)*' ' + e_lines[1])]
        else:
            parse_data.errors += [(file_name, e.line, msg + e_lines[0])]

    return parse_data




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from .graph import Graph

# def ready_edges(g: Graph, edges: Set[int], v_pos: Mapping[int,float]) -> Set[int]:
#     ready = set()
#     for e in edges:
#         src = g.source(e)
#         if all(v in v_pos for v in src):
#             ready.add(e)
#     return ready

def layer_decomp(g: Graph) -> List[List[int]]:
    """Decompose a graph into regular and singular layers

    Returns a list, sorting the edges of `g` into layers. Note that this
    can modify `g` by introducing extra vertices and identity boxes.
    """

    e_layers = []
    v_layer = []
    v_placed = set()


    # first, mark all of the inputs as 'placed' and add dummy edges for any input that is also an output
    outputs = set(g.outputs())
    for v in g.inputs():
        if v in outputs: g.insert_id_after(v)
        v_layer.append(v)
        v_placed.add(v)
    
    new_ids = set()
    edges = set(g.edges())

    # next, place edges in layers
    while len(edges) > 0:
        ready = set()
        for e in edges:
            if all(v in v_placed for v in g.source(e)):
                ready.add(e)
        outputs = set(g.outputs())
        for v in v_layer:
            if v in outputs or any(e not in ready for e in g.out_edges(v)):
                id = g.insert_id_after(v)
                new_ids.add(id)
                ready.add(id)
        
        e_layer = []
        for e in ready:
            src = g.source(e)
            e_layer.append(e)
            edges.discard(e)
        if all(e in new_ids for e in e_layer):
            raise ValueError("Could not make progress. Is graph acyclic?")

        e_layers.append(e_layer)

        if len(edges) > 0:
            v_layer = []
            for e in e_layer:
                for v in g.target(e):
                    v_placed.add(v)
                    v_layer.append(v)

    # finally attempt to minimise crossings by sorting edges according to the ideal positions of their source and
    # target vertices. This is done in a forward (it=0) and backward (it=1) pass.
    for it in range(2):
        rng = range(len(e_layers)) if it == 0 else range(len(e_layers)-1,-1,-1)
        for j in rng:
            inp = [v for e in e_layers[j-1] for v in g.target(e)] if j > 0 else list(g.inputs())
            inp_pos = { v : i/len(inp) for i, v in enumerate(inp) }

            # during the forward pass, ignore the positions of outputs
            if it != 0:
                outp = [v for e in e_layers[j+1] for v in g.source(e)] if j < len(e_layers)-1 else list(g.outputs())
                outp_pos = { v : i/len(outp) for i, v in enumerate(outp) }
            else:
                outp_pos = None

            e_pos = {}
            for e in e_layers[j]:
                src = g.source(e)
                e_pos[e] = sum(inp_pos[v] for v in src)/len(src) if len(src) != 0 else 0

                # during the backward pass, weight the positions of outputs more heavily than inputs
                if outp_pos:
                    tgt = g.target(e)
                    e_pos[e] += 2 * sum(outp_pos[v] for v in tgt)/len(tgt) if len(tgt) != 0 else 0

            list.sort(e_layers[j], key=lambda e: e_pos[e])

    return e_layers

def perm_to_s(perm: List[int]) -> str:
    if len(perm) == 1:
        return 'id'
    elif len(perm) == 2:
        return 'sw'
    else:
        return 'sw' + str(perm)

def split_perm(perm: List[int]) -> List[List[int]]:
    """Split a permutation into a tensor product of independent permuations
    """
    perms = []
    rest = perm

    while rest != []:
        m = 0
        for i, x in enumerate(rest):
            m = max(x, m)
            if m <= i:
                perms.append(rest[:i+1])
                rest = [y-(i+1) for y in rest[i+1:]]
                break
    return perms


def graph_to_term(g: Graph) -> str:
    """Convert a graph to a term

    This currently only works for monogamous acyclic graphs (and hence symmetric monoidal
    terms).
    """

    g = g.copy()
    e_layers = layer_decomp(g)

    in_layer = list(g.inputs())

    # build a list of terms to be sequentially composed
    seq = []
    for i in range(len(e_layers)):
        # compute the permutation from the current vertex layer to inputs of the edge layer
        v_pos = { v : j for j,v in enumerate(in_layer) }
        out_layer = [v for e in e_layers[i] for v in g.source(e)]
        v_perm = [v_pos[v] for v in out_layer]

        # append it as a layer of swap maps
        if v_perm != list(range(len(v_perm))):
            perms = split_perm(v_perm)
            seq.append(' * '.join([perm_to_s(p) for p in perms]))

        # append the parallel composition of the current edge layer
        par = [str(g.edge_data(e).value) for e in e_layers[i]]
        seq.append(' * '.join(par))

        # compute the permutation from the outputs of the edge layer to the next vertex layer
        in_layer = [v for e in e_layers[i] for v in g.target(e)]


    # compute the permutation from the final vertex layer to the outputs
    v_pos = { v : j for j,v in enumerate(in_layer) }
    out_layer = list(g.outputs())
    v_perm = [v_pos[v] for v in out_layer]

    # append it as a layer of swap maps
    if v_perm != list(range(len(v_perm))):
        perms = split_perm(v_perm)
        seq.append(' * '.join([perm_to_s(p) for p in perms]))

    return ' ; '.join(seq)




#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import sys
from PySide6.QtGui import QColor, QPalette
from PySide6.QtWidgets import QApplication

from .colors import apply_theme

from .mainwindow import MainWindow


cat_macchiato_p = {
  'rosewater':  '#f4dbd6',
  'flamingo':   '#f0c6c6',
  'pink':       '#f5bde6',
  'mauve':      '#c6a0f6',
  'red':        '#ed8796',
  'maroon':     '#ee99a0',
  'peach':      '#f5a97f',
  'yellow':     '#eed49f',
  'green':      '#a6da95',
  'teal':       '#8bd5ca',
  'sky':        '#91d7e3',
  'sapphire':   '#7dc4e4',
  'blue':       '#8aadf4',
  'lavender':   '#b7bdf8',
  'text':       '#cad3f5',
  'subtext1':   '#b8c0e0',
  'subtext0':   '#a5adcb',
  'overlay2':   '#939ab7',
  'overlay1':   '#8087a2',
  'overlay0':   '#6e738d',
  'surface2':   '#5b6078',
  'surface1':   '#494d64',
  'surface0':   '#363a4f',
  'base':       '#24273a',
  'mantle':     '#1e2030',
  'crust':      '#181926',
}

catppuccin_macchiato = {
  'bg': cat_macchiato_p['base'],
  'fg': cat_macchiato_p['text'],
  'fg_bright': cat_macchiato_p['lavender'],
  'fg_dim': cat_macchiato_p['overlay1'],
  'fg_good': cat_macchiato_p['green'],
  'fg_bad': cat_macchiato_p['red'],
  'bg_alt': cat_macchiato_p['crust'],
  'bg_button': cat_macchiato_p['mantle'],
  'fg_button': cat_macchiato_p['text'],
  'fg_link': cat_macchiato_p['blue'],
  'bg_highlight': cat_macchiato_p['blue'],
  'fg_highlight': cat_macchiato_p['crust'],
}

theme = catppuccin_macchiato

class Chyp(QApplication):
    """The main chyp application

    ...
    """

    def __init__(self) -> None:
        super().__init__(sys.argv)
        self.setApplicationName('chyp')
        self.setDesktopFileName("chyp")
        apply_theme()
        self.main_window = MainWindow()
        self.lastWindowClosed.connect(self.quit)


def main() -> None:
    """Main entry point for chyp"""

    chyp = Chyp()
    chyp.exec_()



#     chyp - An interactive theorem prover for string diagrams 
#     Copyright (C) 2022 - Aleks Kissinger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import List
from PySide6.QtCore import QDir, QFileInfo, QSettings, Signal
from PySide6.QtGui import QFont, QTextDocument
from PySide6.QtWidgets import QFileDialog, QMessageBox, QPlainTextDocumentLayout, QWidget

from .highlighter import ChypHighlighter

class ChypDocument(QTextDocument):
    fileNameChanged = Signal()
    # documentReplaced = Signal()

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.parent_widget = parent
        self.setDefaultFont(QFont("monospace", 14))
        self.setDocumentLayout(QPlainTextDocumentLayout(self))
        self.highlighter = ChypHighlighter(self)
        self.file_name = ''
        
    def confirm_close(self) -> bool:
        if self.isModified():
            text = "Do you wish to save changes to {}?".format(
                    self.file_name if self.file_name != '' else 'Untitled')
            res = QMessageBox.question(self.parent_widget,
                                       "Save changes",
                                       text,
                                       QMessageBox.StandardButton.Yes |
                                         QMessageBox.StandardButton.No |
                                         QMessageBox.StandardButton.Cancel,
                                       QMessageBox.StandardButton.Yes)

            if res == QMessageBox.StandardButton.Yes:
                return self.save()
            elif res == QMessageBox.StandardButton.No:
                return True
            else:
                return False
        else:
            return True

    def add_to_recent_files(self, file_name: str) -> None:
        conf = QSettings('chyp', 'chyp')
        o = conf.value('recent_files', [])
        recent_files: List[str] = o if isinstance(o, list) else [] 
        recent_files = [f for f in recent_files if f != file_name]
        recent_files.insert(0, file_name)
        recent_files = recent_files[:10]
        conf.setValue('recent_files', recent_files)

    def open(self, file_name: str) -> None:
        self.file_name = file_name
        with open(file_name) as f:
            self.setPlainText(f.read())
        self.add_to_recent_files(self.file_name)
        self.setModified(False)
        self.fileNameChanged.emit()

    def save(self) -> bool:
        if self.file_name:
            with open(self.file_name, 'w') as f:
                f.write(self.toPlainText())
            self.add_to_recent_files(self.file_name)
            self.setModified(False)
            self.fileNameChanged.emit()
            return True
        else:
            return self.save_as()

    def save_as(self) -> bool:
        conf = QSettings('chyp', 'chyp')
        o = conf.value('last_dir')
        last_dir = o if isinstance(o, str) else QDir.home().absolutePath()

        # open save-as box. note we confirm overwrite manually because the GTK/linux native
        # dialog doesn't do this correctly.
        file_name, _ = QFileDialog.getSaveFileName(self.parent_widget,
                                                   "Save File",
                                                   last_dir,
                                                   'chyp files (*.chyp)',
                                                   'chyp files (*.chyp)',
                                                   QFileDialog.Option.DontConfirmOverwrite)
        if file_name:
            if QFileInfo(file_name).suffix() == '':
                file_name += '.chyp'

            fi = QFileInfo(file_name)
            if fi.exists():
                res = QMessageBox.question(self.parent_widget,
                                           "Confirm overwrite",
                                           "File {} exists, do you wish to overwrite it?".format(fi.fileName()))
                if res == QMessageBox.StandardButton.No:
                    return False
            self.file_name = file_name
            conf.setValue('last_dir', QFileInfo(file_name).absolutePath())
            return self.save()
        else:
            return False


